import os
import warnings
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    Iterable,
    List,
    Optional,
    Tuple,
    Union,
)

from spacy.tokens import Doc
from typing_extensions import Literal

from edsnlp import registry
from edsnlp.core import PipelineProtocol
from edsnlp.data.converters import DocToMarkupConverter, MarkupToDocConverter
from edsnlp.pipes.base import BaseNERComponent
from edsnlp.utils.fuzzy_alignment import align
from edsnlp.utils.span_getters import SpanGetterArg, SpanSetterArg, get_spans

from ..async_worker import AsyncRequestWorker

# TODO: find a good API for entity attribute extraction too since xml/md converters
#       support attributes.


@registry.factory.register("eds.llm_markup_extractor")
class LlmMarkupExtractor(BaseNERComponent):
    r'''
    The `eds.llm_markup_extractor` component extracts entities using a
    Large Language Model (LLM) by prompting it to annotate the text with a markup
    format (XML or Markdown). The component can be configured with a set of labels
    to extract, and can be provided with few-shot examples to improve performance.

    In practice, along with a system prompt that describes the allowed labels,
    annotation format and few-shot examples (if any), the component sends documents
    to an LLM like:
    ```
    La patient a une néphropathie diabétique.
    ```

    and expects in return the same text annotated with the entities, for instance:
    ```html
    La patient a une <diag>néphropathie diabétique</diag>.
    ```

    which is then parsed to extract the entities. This approach is close to the one
    of [@naguib_2024] but supports various markup formats and multi label
    prompts. Lookup their paper for more details on the prompting strategies and
    performance.

    !!! warning "Experimental"

        This component is experimental. The API and behavior may change in future
        versions. Make sure to pin your `edsnlp` version if you use it in a project.

    !!! note "Dependencies"

        This component requires several dependencies. Run the following command to
        install them:
        ```bash { data-md-color-scheme="slate" }
        pip install openai bm25s Stemmer
        ```
        We recommend even to add them to your `pyproject.toml` or `requirements.txt`.

    Examples
    --------
    If your data is sensitive, we recommend you to use a self-hosted
    model with an OpenAI-compatible API, such as
    [vLLM](https://github.com/vllm-project/vllm).

    Start a server with the model of your choice:

    ```bash { data-md-color-scheme="slate" }
    python -m vllm.entrypoints.openai.api_server \
       --model mistral-small-24b-instruct-2501 \
       --port 8080 \
       --enable-prefix-caching
    ```

    You can then use the `llm_markup_extractor` component as follows:

    <!-- blacken-docs:off -->

    ```python { .no-check }
    import edsnlp, edsnlp.pipes as eds

    prompt = """
    You are a XML-based extraction assistant.
    For every piece of text the user provides, you will rewrite the full text
    word for word, adding XML tags around the relevant pieces of information.

    You must follow these rules strictly:
    - You must only use the provided tags. Do not invent new tags.
    - You must follow the original text exactly: do not alter it, only add tags.
    - You must always close every tag you open.
    - If a piece of text does not contain any of the information to extract, you must return the text unchanged, without any tags.
    - Be consistent in your answers, similar queries must lead to similar answers, do not try to fix your prior answers.
    - Do not add any comment or explanation, just write the text with tags.

    Example with an <noun_group> tag:
    User query: "This is a sample document."
    Assistant answer: "This is <noun_group>a sample document</noun_group>."

    The tags to use are the following:
    - <diag>: A medical diagnosis
    - <treat>: A medical treatment
    """.strip()

    # EDS-NLP util to create documents from Markdown or XML markup.
    # This has nothing to do with the LLM component itself.
    conv = edsnlp.data.converters.MarkupToDocConverter(preset="xml")
    train_docs = [  # (1)!
        conv("Le patient a une <diag>pneumonie</diag>."),
        conv("On prescrit l'<treat>antibiothérapie</treat>."),
        # ... add more examples if you can
    ]

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.llm_markup_extractor(
            # OpenAI-compatible API like the local vLLM server above
            api_url="http://localhost:8080/v1",
            model="my-custom-model",
            examples=train_docs,
            # Apply the model to each sentence separately
            context_getter="sents",
            # String or function that returns a list of messages (see below)
            prompt=prompt,
            use_retriever=True,
            # For each request, show the model the closest example
            max_few_shot_examples=1,
            # Up to 5 requests in parallel
            max_concurrent_requests=5,
        )
    )
    doc = nlp("Le patient souffre de tuberculose. On débute une antibiothérapie.")
    print([(ent.text, ent.label_) for ent in doc.ents])
    # Out: [('tuberculose', 'diag'), ('antibiothérapie', 'treat')]
    ```

    1. You could also use [EDS-NLP's data API](/data/)
       ```python
       import edsnlp

       train_docs = edsnlp.data.from_iterable(
           [
               "Le patient a une <diag>pneumonie</diag>.",
               "On prescrit l'<treat>antibiothérapie</treat>.",
           ],
           converter="markup",
           preset="xml",
       )
       ```

    <!-- blacken-docs:on -->

    You can also control the prompt more finely by providing a callable
    instead of a string. For instance, let's put all few-shot examples
    in the system message, and the actual user query in a single user message:

    ```python { .no-check }
    def prompt(doc_text, examples):
        system_content = (
            "You are a XML-based extraction assistant.\n"
            "Here are some examples of what's expected:\n"
        )
        for ex_text, ex_markup in examples:
            system_content += f"- User: {ex_text}\n"
            system_content += f"  Bot answer: {ex_markup}\n"
        return [
            {"role": "system", "content": system_content},
            {"role": "user", "content": doc_text},
        ]
    ```

    Parameters
    ----------
    nlp : PipelineProtocol
        Pipeline object.
    name : str
        Component name.
    api_url : str
        The base URL of the OpenAI-compatible API.
        You must explicitly provide this to avoid leaking requests to
        the public OpenAI API. Should you work with sensitive data,
        consider using a self-hosted model.
    model : str
        The model name to use. Must be available on the API server.
    markup_mode : Literal["xml", "md"]
        The markup format to use when formatting the few-shot examples and
        parsing the model's output. Either "xml" (default) or "md" (Markdown).
        Make sure the prompt template matches the chosen format.
    prompt : Union[str, Callable[[str, List[Tuple[str, str]]], List[Dict[str, str]]]]
        The prompt is the main way to control the model's behavior.
        It can be either:

        - A string, which will be used as a system prompt.
          Few-shot examples (if any) will be provided as user/assistant
          messages before the actual user query.
        - A callable that takes two arguments:

            * `doc_text`: the text of the document to process
            and returns a list of messages in the format expected by the
            OpenAI chat completions API.
            * `examples`: a list of few-shot examples, each being a tuple
              of (text, markup annotated text)
    examples : Optional[Iterable[Doc]]
        Few-shot examples to provide to the model. The more the better, but
        the total number of tokens in the prompt must be less than the model's
        context size. If `use_retriever` is set to `True`, the most relevant
        examples will be selected automatically.
    max_few_shot_examples : int
        The maximum number of few-shot examples to provide to the model.
        Default to -1 (all examples).
    use_retriever : Optional[bool]
        Whether to use a retriever to select the most relevant few-shot examples.
        If `None` (default), it will be set to `True` if `max_few_shot_examples` is
        greater than 0 and the number of examples is greater than `max_few_shot_examples`.
        If set to `False`, the first `max_few_shot_examples` will be used.
    context_getter : Optional[SpanGetterArg]
        This parameter controls the contexts given to the model for each request.
        It can be used to split the document into smaller chunks, for instance
        sentences by setting `context_getter="sents"`, or process just a part of the
        document, for instance with `context_getter={"sections": "conclusion"}`.
        If `None` (default), the whole document is processed in a single request.
    span_setter : SpanSetterArg
        On which span group (`doc.spans[...]` or `doc.ents`) to set the extracted
        entities.
    span_getter : Optional[SpanGetterArg]
        From which span group (`doc.spans[...]` or `doc.ents`) to get the spans
        to annotate from few-shot examples. Default to the same as `span_setter`.
    seed : Optional[int]
        Optional seed forwarded to the API.
    max_concurrent_requests : int
        Maximum number of concurrent span requests per document.
    api_kwargs : Dict[str, Any]
        Extra keyword arguments forwarded to `chat.completions.create`.
    on_error : Literal["raise", "warn"]
        Error handling strategy. If `"raise"`, exceptions are raised. If `"warn"`,
        exceptions are logged as warnings and processing continues.

    Authors and citation
    --------------------
    The `eds.llm_markup_extractor` component was developed by AP-HP's Data Science
    team.
    '''  # noqa: E501

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: str = "llm_markup_extractor",
        *,
        api_url: str,
        model: str,
        prompt: Union[
            str, Callable[[str, List[Tuple[str, str]]], List[Dict[str, str]]]
        ],
        markup_mode: Literal["xml", "md"] = "xml",
        examples: Iterable[Doc] = (),
        max_few_shot_examples: int = -1,
        use_retriever: Optional[bool] = None,
        context_getter: SpanGetterArg = None,
        span_setter: SpanSetterArg = {"ents": True},
        span_getter: Optional[SpanGetterArg] = None,
        seed: Optional[int] = None,
        max_concurrent_requests: int = 1,
        api_kwargs: Optional[Dict[str, Any]] = None,
        on_error: Literal["raise", "warn"] = "raise",
    ):
        import openai

        self.lang = nlp.lang
        self.api_url = api_url
        self.model = model
        if span_getter is None:
            span_getter = span_setter
        self.context_getter = context_getter
        self.markup_to_doc = MarkupToDocConverter(
            preset=markup_mode,
            span_setter=span_setter,
        )
        self.doc_to_markup = DocToMarkupConverter(
            preset=markup_mode,
            span_getter=span_getter,
        )
        self.examples = [(doc.text, self.doc_to_markup(doc)) for doc in examples]
        self.max_few_shot_examples = max_few_shot_examples
        self.span_setter = span_setter
        # Double check just in case, but confit should have caught this
        assert api_url is not None, "api_url must be provided"
        api_key = os.getenv("OPENAI_API_KEY", "")
        self.client = openai.OpenAI(base_url=self.api_url, api_key=api_key)
        self.async_client = openai.AsyncOpenAI(base_url=self.api_url, api_key=api_key)
        self.prompt = prompt
        self.api_kwargs = api_kwargs or {}
        self.max_concurrent_requests = max_concurrent_requests
        self.on_error = on_error
        self.seed = seed
        self.retriever = None
        if self.max_few_shot_examples > 0 and use_retriever is not False:
            self.build_few_shot_retriever_(self.examples)
        super().__init__(nlp=nlp, name=name, span_setter=span_setter)

    def _handle_err(self, msg):
        if self.on_error == "raise":
            raise RuntimeError(msg)
        else:
            warnings.warn(msg)

    def set_extensions(self) -> None:
        super().set_extensions()

    def __call__(self, doc):
        return self.process(doc)

    def apply_markup_to_doc_(self, doclike: Any, markup_answer: str):
        """
        Apply the markup answer to the doclike object.
        """
        res_text, res_ents = self.markup_to_doc._parse(markup_answer)
        res_text = res_text.rstrip()
        stripped_text = doclike.text.rstrip()

        if stripped_text != res_text:
            ents = [
                {"fragments": [{"begin": s, "end": e}], "label": lab, "attributes": a}
                for s, e, lab, a in res_ents
            ]
            aligned = align(
                {"text": res_text, "entities": ents},
                {"text": stripped_text, "entities": []},
            )
            res_ents = [
                (f["begin"], f["end"], e["label"], e["attributes"])
                for e in aligned["doc"]["entities"]
                for f in e["fragments"]
            ]

        spans = []
        for start, end, label, attrs in res_ents:
            span = doclike.char_span(start, end, label=label, alignment_mode="expand")
            if span is None:
                continue
            spans.append(span)

        doc = doclike.doc if hasattr(doclike, "doc") else doclike
        self.set_spans(doc, spans)

    def build_few_shot_retriever_(self, samples):
        # TODO: put in a new edsnlp retrievers module?
        import bm25s
        import Stemmer

        lang = {"eds": "french"}.get(self.lang, self.lang)
        stemmer = Stemmer.Stemmer(lang)
        texts = [s[0] for s in samples]
        corpus = bm25s.tokenize(texts, stemmer=stemmer, stopwords=lang)
        retriever = bm25s.BM25()
        retriever.index(corpus)
        self.retriever = retriever
        self.retriever.stemmer = stemmer

    def build_prompt(self, doc):
        import bm25s

        few_shot_examples = []
        if self.retriever is not None:
            closest_texts_indices, scores = self.retriever.retrieve(
                bm25s.tokenize(
                    doc.text,
                    stemmer=self.retriever.stemmer,
                    show_progress=False,
                ),
                k=self.max_few_shot_examples,
                show_progress=False,
            )
            for i in closest_texts_indices[0][: self.max_few_shot_examples]:
                few_shot_examples.append(self.examples[i])
            few_shot_examples = few_shot_examples[::-1]  # reverse to have closest last
        else:
            few_shot_examples = self.examples[: self.max_few_shot_examples]

        if isinstance(self.prompt, str):
            messages = [{"role": "system", "content": self.prompt}]
            for ex_text, ex_markup in few_shot_examples:
                messages.append({"role": "user", "content": ex_text})
                messages.append({"role": "assistant", "content": ex_markup})
            messages.append({"role": "user", "content": doc.text})
        else:
            messages = self.prompt(doc.text, few_shot_examples)
        return messages

    def process(self, doc):
        """
        Handle a single doc
        """
        for ctx in get_spans(doc, self.context_getter):
            messages = self.build_prompt(ctx)
            markup_answer = self._llm_request_sync(messages)
            self.apply_markup_to_doc_(ctx, markup_answer)

        return doc

    def pipe(self, docs: Iterable[Doc]) -> Iterable[Doc]:
        """
        Extract entities concurrently, but yield results in the same order
        as the input `docs`. Up to `max_concurrent_requests` documents are
        processed in parallel.

        Parameters
        ----------
        docs: Iterable[Doc]
            Documents to process

        Yields
        ------
        Doc
            Processed documents in the original input order.
        """
        if self.max_concurrent_requests <= 1:  # pragma: no cover
            for ctx in docs:
                yield self.process(ctx)
            return

        worker = AsyncRequestWorker.instance()

        # Documents sent to the worker, waiting for results
        pending_docs: Dict[int, Doc] = {}
        # Documents already processed, waiting to be yielded in order
        buffer: Dict[int, Doc] = {}
        next_to_yield = 0
        in_flight: Dict[int, int] = {}

        ctx_iter = enumerate(
            ctx for doc in docs for ctx in get_spans(doc, self.context_getter)
        )

        for _ in range(self.max_concurrent_requests):
            try:
                i, ctx = next(ctx_iter)
            except StopIteration:
                break
            messages = self.build_prompt(ctx)
            task_id = worker.submit(self._llm_request_coro(messages))
            in_flight[task_id] = i
            pending_docs[i] = ctx

        while in_flight:
            done_task_id = worker.wait_for_any(in_flight.keys())
            result = worker.pop_result(done_task_id)
            i = in_flight.pop(done_task_id)
            ctx = pending_docs.pop(i)

            if result is None:
                buffer[i] = ctx
            else:
                res, err = result
                if err is not None:
                    self._handle_err(
                        f"[llm_markup_extractor] failed for document #{i}: {err!r}"
                    )
                    buffer[i] = ctx
                else:
                    try:
                        self.apply_markup_to_doc_(ctx, str(res))
                    except Exception as e:  # pragma: no cover
                        import traceback

                        traceback.print_exc()
                        self._handle_err(
                            f"[llm_markup_extractor] "
                            f"failed to parse result for document #{i}: {e!r} in "
                            f"{res!r}"
                        )
                    buffer[i] = ctx

            while True:
                try:
                    if len(in_flight) >= self.max_concurrent_requests:
                        break
                    i2, d2 = next(ctx_iter)
                except StopIteration:
                    break
                messages2 = self.build_prompt(d2)
                task_id2 = worker.submit(self._llm_request_coro(messages2))
                in_flight[task_id2] = i2
                pending_docs[i2] = d2

            while next_to_yield in buffer:
                yield buffer.pop(next_to_yield)
                next_to_yield += 1

        while next_to_yield in buffer:  # pragma: no cover
            yield buffer.pop(next_to_yield)
            next_to_yield += 1

    def _llm_request_sync(self, messages) -> str:
        response = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
            seed=self.seed,
            **self.api_kwargs,
        )
        return response.choices[0].message.content

    def _llm_request_coro(self, messages) -> Coroutine[Any, Any, str]:
        async def _coro():
            response = await self.async_client.chat.completions.create(
                model=self.model,
                messages=messages,
                seed=self.seed,
                **self.api_kwargs,
            )
            return response.choices[0].message.content

        return _coro()



from __future__ import annotations

import logging
from typing import Any, Dict, Iterable, List, Optional, Sequence, Set, Tuple, cast

import foldedtensor as ft
import numpy as np
import torch
import torch.nn.functional as F
from spacy.tokens import Doc, Span
from typing_extensions import Literal

from edsnlp.core import PipelineProtocol
from edsnlp.core.torch_component import BatchInput, BatchOutput, TorchComponent
from edsnlp.pipes.trainable.embeddings.typing import WordEmbeddingComponent
from edsnlp.utils.span_getters import SpanGetterArg, get_spans

logger = logging.getLogger(__name__)


# ===============================================================
def chuliu_edmonds_one_root(scores: np.ndarray) -> np.ndarray:
    """
    Shamelessly copied from
    https://github.com/hopsparser/hopsparser/blob/main/hopsparser/mst.py#L63
    All credits, Loic Grobol at Université Paris Nanterre, France, the
    original author of this implementation. Find the license of the hopsparser software
    below:

    Copyright 2020 Benoît Crabbé benoit.crabbe@linguist.univ-paris-diderot.fr

    Permission is hereby granted, free of charge, to any person obtaining a copy of
    this software and associated documentation files (the "Software"), to deal in the
    Software without restriction, including without limitation the rights to use,
    copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
    Software, and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
    HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
    OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
    SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

    ---

    Repeatedly Use the Chu‑Liu/Edmonds algorithm to find a maximum spanning
    dependency tree from the weight matrix of a rooted weighted directed graph.

    **ATTENTION: this modifies `scores` in place.**

    ## Input

    - `scores`: A 2d numeric array such that `scores[i][j]` is the weight
    of the `$j→i$` edge in the graph and the 0-th node is the root.

    ## Output

    - `tree`: A 1d integer array such that `tree[i]` is the head of the `i`-th node
    """

    # FIXME: we don't actually need this in CLE: we only need one critical cycle
    def tarjan(tree: np.ndarray) -> List[np.ndarray]:
        """Use Tarjan's SCC algorithm to find cycles in a tree

        ## Input

        - `tree`: A 1d integer array such that `tree[i]` is the head of
           the `i`-th node

        ## Output

        - `cycles`: A list of 1d bool arrays such that `cycles[i][j]` is
           true iff the `j`-th node of
          `tree` is in the `i`-th cycle
        """
        indices = -np.ones_like(tree)
        lowlinks = -np.ones_like(tree)
        onstack = np.zeros_like(tree, dtype=bool)
        stack = list()
        # I think this is in a list to be able to mutate it in the closure, even
        # though `nonlocal` exists
        _index = [0]
        cycles = []

        def strong_connect(i):
            _index[0] += 1
            index = _index[-1]  # `_index` is of length 1 so this is also `_index[0]`???
            indices[i] = lowlinks[i] = index - 1
            stack.append(i)
            onstack[i] = True
            dependents = np.where(np.equal(tree, i))[0]
            for j in dependents:
                if indices[j] == -1:
                    strong_connect(j)
                    lowlinks[i] = min(lowlinks[i], lowlinks[j])
                elif onstack[j]:
                    lowlinks[i] = min(lowlinks[i], indices[j])

            # There's a cycle!
            if lowlinks[i] == indices[i]:
                cycle = np.zeros_like(indices, dtype=bool)
                while stack[-1] != i:
                    j = stack.pop()
                    onstack[j] = False
                    cycle[j] = True
                stack.pop()
                onstack[i] = False
                cycle[i] = True
                if cycle.sum() > 1:
                    cycles.append(cycle)
            return

        # -------------------------------------------------------------
        for i in range(len(tree)):
            if indices[i] == -1:
                strong_connect(i)
        return cycles

    # TODO: split out a `contraction` function to make this more readable
    def chuliu_edmonds(scores: np.ndarray) -> np.ndarray:
        """Use the Chu‑Liu/Edmonds algorithm to find a maximum spanning
        arborescence from the weight matrix of a rooted weighted directed
        graph

        ## Input

        - `scores`: A 2d numeric array such that `scores[i][j]` is the
          weight of the `$j→i$` edge in the graph and the 0-th node is the root.

        ## Output

        - `tree`: A 1d integer array such that `tree[i]` is the head of the `i`-th node
        """
        np.fill_diagonal(scores, -float("inf"))  # prevent self-loops
        scores[0] = -float("inf")
        scores[0, 0] = 0
        tree = cast(np.ndarray, np.argmax(scores, axis=1))
        cycles = tarjan(tree)
        if not cycles:
            return tree
        else:
            # t = len(tree); c = len(cycle); n = len(noncycle)
            # locations of cycle; (t) in [0,1]
            cycle = cycles.pop()
            # indices of cycle in original tree; (c) in t
            cycle_locs = np.where(cycle)[0]
            # heads of cycle in original tree; (c) in t
            cycle_subtree = tree[cycle]
            # scores of cycle in original tree; (c) in R
            cycle_scores = scores[cycle, cycle_subtree]
            # total score of cycle; () in R
            total_cycle_score = cycle_scores.sum()

            # locations of noncycle; (t) in [0,1]
            noncycle = np.logical_not(cycle)
            # indices of noncycle in original tree; (n) in t
            noncycle_locs = np.where(noncycle)[0]

            # scores of cycle's potential heads; (c x n) - (c) + () -> (n x c) in R
            metanode_head_scores = (
                scores[cycle][:, noncycle]
                - cycle_scores[:, np.newaxis]
                + total_cycle_score
            )
            # scores of cycle's potential dependents; (n x c) in R
            metanode_dep_scores = scores[noncycle][:, cycle]
            # best noncycle head for each cycle dependent; (n) in c
            metanode_heads = np.argmax(metanode_head_scores, axis=0)
            # best cycle head for each noncycle dependent; (n) in c
            metanode_deps = np.argmax(metanode_dep_scores, axis=1)

            # scores of noncycle graph; (n x n) in R
            subscores = scores[noncycle][:, noncycle]
            # expand to make space for the metanode (n+1 x n+1) in R
            subscores = np.pad(subscores, ((0, 1), (0, 1)), "constant")
            # set the contracted graph scores of cycle's potential
            # heads; (c x n)[:, (n) in n] in R -> (n) in R
            subscores[-1, :-1] = metanode_head_scores[
                metanode_heads, np.arange(len(noncycle_locs))
            ]
            # set the contracted graph scores of cycle's potential
            # dependents; (n x c)[(n) in n] in R-> (n) in R
            subscores[:-1, -1] = metanode_dep_scores[
                np.arange(len(noncycle_locs)), metanode_deps
            ]

            # MST with contraction; (n+1) in n+1
            contracted_tree = chuliu_edmonds(subscores)
            # head of the cycle; () in n
            cycle_head = contracted_tree[-1]
            # fixed tree: (n) in n+1
            contracted_tree = contracted_tree[:-1]
            # initialize new tree; (t) in 0
            new_tree = -np.ones_like(tree)
            # fixed tree with no heads coming from the cycle: (n) in [0,1]
            contracted_subtree = contracted_tree < len(contracted_tree)
            # add the nodes to the new tree (t)[(n)[(n) in [0,1]] in t]
            # in t = (n)[(n)[(n) in [0,1]] in n] in t
            new_tree[noncycle_locs[contracted_subtree]] = noncycle_locs[
                contracted_tree[contracted_subtree]
            ]
            # fixed tree with heads coming from the cycle: (n) in [0,1]
            contracted_subtree = np.logical_not(contracted_subtree)
            # add the nodes to the tree (t)[(n)[(n) in [0,1]] in t]
            # in t = (c)[(n)[(n) in [0,1]] in c] in t
            new_tree[noncycle_locs[contracted_subtree]] = cycle_locs[
                metanode_deps[contracted_subtree]
            ]
            # add the old cycle to the tree; (t)[(c) in t] in t = (t)[(c) in t] in t
            new_tree[cycle_locs] = tree[cycle_locs]
            # root of the cycle; (n)[() in n] in c = () in c
            cycle_root = metanode_heads[cycle_head]
            # add the root of the cycle to the new
            # tree; (t)[(c)[() in c] in t] = (c)[() in c]
            new_tree[cycle_locs[cycle_root]] = noncycle_locs[cycle_head]
            return new_tree

    scores = scores.astype(np.float64)
    tree = chuliu_edmonds(scores)
    roots_to_try = np.where(np.equal(tree[1:], 0))[0] + 1

    # PW: small change here (<= instead of ==) to avoid crashes in pathological cases
    if len(roots_to_try) <= 1:
        return tree

    # -------------------------------------------------------------
    def set_root(scores: np.ndarray, root: int) -> Tuple[np.ndarray, np.ndarray]:
        """Force the `root`-th node to be the only node under the root by overwriting
        the weights of the other children of the root."""
        root_score = scores[root, 0]
        scores = np.array(scores)
        scores[1:, 0] = -float("inf")
        scores[root] = -float("inf")
        scores[root, 0] = 0
        return scores, root_score

    # We find the maximum spanning dependency tree by trying every possible root
    best_score, best_tree = -np.inf, None  # This is what's causing it to crash
    for root in roots_to_try:
        _scores, root_score = set_root(scores, root)
        _tree = chuliu_edmonds(_scores)
        tree_probs = _scores[np.arange(len(_scores)), _tree]
        tree_score = (
            (tree_probs).sum() + (root_score)
            if (tree_probs > -np.inf).all()
            else -np.inf
        )
        if tree_score > best_score:
            best_score = tree_score
            best_tree = _tree

    assert best_tree is not None
    return best_tree


class MLP(torch.nn.Module):
    def __init__(
        self, input_dim: int, hidden_dim: int, output_dim: int, dropout_p: float = 0.0
    ):
        super().__init__()
        self.hidden = torch.nn.Linear(input_dim, hidden_dim)
        self.output = torch.nn.Linear(hidden_dim, output_dim)
        self.dropout = torch.nn.Dropout(dropout_p)

    def forward(self, x):
        x = self.hidden(x)
        x = F.relu(x)
        x = self.dropout(x)
        x = self.output(x)
        return x


class BiAffine(torch.nn.Module):
    def __init__(
        self,
        in_features: int,
        out_features,
    ):
        super().__init__()
        self.bilinear = torch.nn.Bilinear(
            in_features, in_features, out_features, bias=True
        )
        self.head_linear = torch.nn.Linear(in_features, out_features, bias=False)
        self.tail_linear = torch.nn.Linear(in_features, out_features, bias=False)

    def forward(self, u, v):
        scores = torch.einsum("bux,bvy,lxy->buvl", u, v, self.bilinear.weight)
        scores = scores + self.head_linear(u).unsqueeze(2)
        scores = scores + self.tail_linear(v).unsqueeze(1)
        scores = scores + self.bilinear.bias
        return scores


class TrainableBiaffineDependencyParser(
    TorchComponent[BatchOutput, BatchInput],
):
    """
    The `eds.biaffine_dep_parser` component is a trainable dependency parser
    based on a biaffine model ([@dozat2017deepbiaffineattentionneural]). For each
    token, the model predicts a score for each possible head in the document, and a
    score for each possible label for each head. The results are then decoded either
    greedily by picking the best scoring head for each token independently, or
    holistically by computing the Maximum Spanning Tree (MST) over the graph of
    token → head scores.

    !!! warning "Experimental"

        This component is experimental. In particular, it expects the input to be
        sentences and not full documents, as it has not been optimized for memory
        efficiency yet and computed the full matrix of scores for all pairs of tokens
        in a document.

        At the moment, it is mostly used for benchmarking and research purposes.

    Examples
    --------
    ```{ .python }
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(
        eds.biaffine_dep_parser(
            embedding=eds.transformer(model="prajjwal1/bert-tiny"),
            hidden_size=128,
            dropout_p=0.1,
            # labels unset, will be inferred from the data in `post_init`
            decoding_mode="mst",
        ),
        name="dep_parser"
    )
    ```

    Dependency parsers are typically trained on CoNLL-formatted
    [Universal Dependencies corpora](https://universaldependencies.org/#download),
    which you can load using the [`edsnlp.data.read_conll`][edsnlp.data.read_conll]
    function.

    To train the model, you can adapt the the
    [Training NER tutorial](/tutorials/training-ner).

    Parameters
    ----------
    nlp: Optional[PipelineProtocol]
        The pipeline object
    name: str
        Name of the component
    embedding: WordEmbeddingComponent
        The word embedding component
    context_getter: Optional[SpanGetterArg]
        What context to use when computing the span embeddings (defaults to the whole
        document). For example `{"section": "conclusion"}` to predict dependencies
        in the conclusion section of documents.
    use_attrs: Optional[List[str]]
        The attributes to use as features for the model (ex. `["pos_"]` to use the POS
        tag). By default, no attributes are used.

        Note that if you train a model with attributes, you will need to provide the
        same attributes during inference, and the model might not work well if the
        attributes were not annotated accurately on the test data.
    attr_size: int
        The size of the attribute embeddings.
    hidden_size: int
        The size of the hidden layer in the MLP.
    dropout_p: float
        The dropout probability to use in the MLP.
    labels: List[str]
        The labels to predict. The labels can also be inferred from the data during
        `nlp.post_init(...)`.
    decoding_mode: Literal["greedy", "mst"]
        Whether to decode the dependencies greedily or using the Maximum Spanning Tree
        algorithm.

    Authors and citation
    --------------------
    The `eds.biaffine_dep_parser` trainable pipe was developed by
    AP-HP's Data Science team, and heavily inspired by the implementation of
    [@grobol:hal-03223424]. The biaffine architecture is based on the biaffine parser
    of [@dozat2017deepbiaffineattentionneural].
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol] = None,
        name: str = "biaffine_dep_parser",
        *,
        embedding: WordEmbeddingComponent,
        context_getter: Optional[SpanGetterArg] = None,
        use_attrs: Optional[List[str]] = None,
        attr_size: int = 32,
        hidden_size: int = 128,
        dropout_p: float = 0.0,
        labels: List[str] = ["root"],
        decoding_mode: Literal["greedy", "mst"] = "mst",
    ):
        super().__init__(nlp=nlp, name=name)
        self.embedding = embedding
        self.use_attrs: List[str] = use_attrs or []
        self.labels = list(labels) or []
        self.labels_to_idx = {label: idx for idx, label in enumerate(self.labels)}
        self.context_getter = context_getter
        cat_dim = self.embedding.output_size + len(self.use_attrs) * attr_size
        self.head_mlp = MLP(cat_dim, hidden_size, hidden_size, dropout_p)
        self.tail_mlp = MLP(cat_dim, hidden_size, hidden_size, dropout_p)
        self.arc_biaffine = BiAffine(hidden_size, 1)
        self.lab_biaffine = BiAffine(hidden_size, len(self.labels))
        self.root_embed = torch.nn.Parameter(torch.randn(cat_dim)[None, None, :])
        self.decoding_mode = decoding_mode
        self.attr_vocabs = {attr: [] for attr in self.use_attrs}
        self.attr_to_idx = {attr: {} for attr in self.use_attrs}
        self.attr_embeddings = torch.nn.ModuleDict(
            {
                attr: torch.nn.Embedding(len(vocab), attr_size)
                for attr, vocab in self.attr_vocabs.items()
            }
        )

    def update_labels(self, labels: Sequence[str], attrs: Dict[str, List[str]]):
        old_labs = self.labels if self.labels is not None else ()
        old_index = torch.as_tensor(
            [i for i, lab in enumerate(old_labs) if lab in labels], dtype=torch.long
        )
        new_index = torch.as_tensor(
            [labels.index(lab) for lab in old_labs if lab in labels], dtype=torch.long
        )
        new_biaffine = BiAffine(self.arc_biaffine.bilinear.in1_features, len(labels))
        # fmt: off
        new_biaffine.bilinear.weight.data[new_index] = self.arc_biaffine.bilinear.weight.data[old_index]  # noqa: E501
        new_biaffine.bilinear.bias.data[new_index] = self.arc_biaffine.bilinear.bias.data[old_index]  # noqa: E501
        new_biaffine.head_linear.weight.data[new_index] = self.arc_biaffine.head_linear.weight.data[old_index]  # noqa: E501
        new_biaffine.tail_linear.weight.data[new_index] = self.arc_biaffine.tail_linear.weight.data[old_index]  # noqa: E501
        # fmt: on
        self.lab_biaffine.bilinear.weight.data = new_biaffine.bilinear.weight.data
        self.lab_biaffine.bilinear.bias.data = new_biaffine.bilinear.bias.data
        self.lab_biaffine.head_linear.weight.data = new_biaffine.head_linear.weight.data
        self.lab_biaffine.tail_linear.weight.data = new_biaffine.tail_linear.weight.data
        self.labels = labels
        self.labels_to_idx = {lab: i for i, lab in enumerate(labels)}

        for attr, vals in attrs.items():
            emb = self.attr_embeddings[attr]
            old_vals = (
                self.attr_vocabs[attr] if self.attr_vocabs[attr] is not None else ()
            )
            old_index = torch.as_tensor(
                [i for i, val in enumerate(old_vals) if val in vals], dtype=torch.long
            )
            new_index = torch.as_tensor(
                [vals.index(val) for val in old_vals if val in vals], dtype=torch.long
            )
            new_emb = torch.nn.Embedding(
                len(vals), self.attr_embeddings[attr].weight.size(1)
            )
            new_emb.weight.data[new_index] = emb.weight.data[old_index]
            self.attr_embeddings[attr].weight.data = new_emb.weight.data
            self.attr_vocabs[attr] = vals
            self.attr_to_idx[attr] = {val: i for i, val in enumerate(vals)}

    def post_init(self, gold_data: Iterable[Doc], exclude: Set[str]):
        super().post_init(gold_data, exclude=exclude)
        labels = dict()
        attr_vocabs = {attr: dict() for attr in self.use_attrs}
        for doc in gold_data:
            ctxs = (
                get_spans(doc, self.context_getter) if self.context_getter else [doc[:]]
            )
            for ctx in ctxs:
                for token in ctx:
                    labels[token.dep_] = True
                    for attr in self.use_attrs:
                        attr_vocabs[attr][getattr(token, attr)] = True
        self.update_labels(
            labels=list(labels.keys()),
            attrs={attr: list(v.keys()) for attr, v in attr_vocabs.items()},
        )

    def preprocess(self, doc: Doc, **kwargs) -> Dict[str, Any]:
        ctxs = get_spans(doc, self.context_getter) if self.context_getter else [doc[:]]
        prep = {
            "embedding": self.embedding.preprocess(doc, contexts=ctxs, **kwargs),
            "$contexts": ctxs,
            "stats": {
                "dep_words": 0,
            },
        }
        for attr in self.use_attrs:
            prep[attr] = [
                [self.attr_to_idx[attr].get(getattr(token, attr), 0) for token in ctx]
                for ctx in ctxs
            ]
        for ctx in ctxs:
            prep["stats"]["dep_words"] += len(ctx) + 1
        return prep

    def preprocess_supervised(self, doc: Doc) -> Dict[str, Any]:
        preps = self.preprocess(doc)
        arc_targets = []  # head idx for each token
        lab_targets = []  # arc label idx for each token
        for ctx in preps["$contexts"]:
            ctx_start = ctx.start
            arc_targets.append(
                [
                    0,
                    *(
                        token.head.i - ctx_start + 1 if token.head.i != token.i else 0
                        for token in ctx
                    ),
                ]
            )
            lab_targets.append([0, *(self.labels_to_idx[t.dep_] for t in ctx)])
        return {
            **preps,
            "arc_targets": arc_targets,
            "lab_targets": lab_targets,
        }

    def collate(self, preps: Dict[str, Any]) -> BatchInput:
        collated = {"embedding": self.embedding.collate(preps["embedding"])}
        collated["stats"] = {
            k: sum(v) for k, v in preps["stats"].items() if not k.startswith("__")
        }
        for attr in self.use_attrs:
            collated[attr] = ft.as_folded_tensor(
                preps[attr],
                data_dims=("context", "tail"),
                full_names=("sample", "context", "tail"),
                dtype=torch.long,
            )
        if "arc_targets" in preps:
            collated["arc_targets"] = ft.as_folded_tensor(
                preps["arc_targets"],
                data_dims=("context", "tail"),
                full_names=("sample", "context", "tail"),
                dtype=torch.long,
            )
        if "lab_targets" in preps:
            collated["lab_targets"] = ft.as_folded_tensor(
                preps["lab_targets"],
                data_dims=("context", "tail"),
                full_names=("sample", "context", "tail"),
                dtype=torch.long,
            )

        return collated

    # noinspection SpellCheckingInspection
    def forward(self, batch: BatchInput) -> BatchOutput:
        embeds = self.embedding(batch["embedding"])["embeddings"]
        embeds = embeds.refold("context", "word")
        embeds = torch.cat(
            [
                embeds,
                *(self.attr_embeddings[attr](batch[attr]) for attr in self.use_attrs),
            ],
            dim=-1,
        )

        embeds_with_root = torch.cat(
            [
                self.root_embed.expand(embeds.shape[0], 1, self.root_embed.size(-1)),
                embeds,
            ],
            dim=1,
        )

        tail_embeds = self.tail_mlp(embeds_with_root)  # (contexts, tail=words, dim)
        head_embeds = self.head_mlp(embeds_with_root)  # (contexts, words+root, dim)

        # Scores: (contexts, tail=1+words, head=1+words)
        arc_logits = self.arc_biaffine(tail_embeds, head_embeds).squeeze(-1)
        # Scores: (contexts, tail=1+words, head=1+words, labels)
        lab_logits = self.lab_biaffine(tail_embeds, head_embeds)

        if "arc_targets" in batch:
            num_labels = lab_logits.shape[-1]
            arc_targets = batch["arc_targets"]  # (contexts, tail=1+words) -> head_idx
            lab_targets = batch["lab_targets"]
            # arc_targets: (contexts, tail=1+words) -> head_idx
            flat_arc_logits = (
                # (contexts, tail=1+words, head=1+words)
                arc_logits.masked_fill(~arc_targets.mask[:, None, :], -10000)[
                    # -> (all_flattened_tails_with_root, head=1+words)
                    arc_targets.mask
                ]
            )
            flat_arc_targets = arc_targets[arc_targets.mask]
            arc_loss = (
                F.cross_entropy(
                    # (all_flattened_tails_with_root, head_with_root)
                    flat_arc_logits,
                    flat_arc_targets,
                    reduction="sum",
                )
                / batch["stats"]["dep_words"]
            )
            flat_lab_logits = (
                # lab_logits: (contexts, tail=1+words, head=1+words, labels)
                lab_logits[arc_targets.mask]
                # -> (all_flattened_tails_with_root, head=1+words, labels)
                .gather(1, flat_arc_targets[:, None, None].expand(-1, 1, num_labels))
                # -> (all_flattened_tails_with_root, 1, labels)
                .squeeze(1)
                # -> (all_flattened_tails_with_root, labels)
            )
            # TODO, directly in collate
            flat_lab_targets = (
                lab_targets[
                    # (contexts, tail=1+words) -> label
                    arc_targets.mask
                ]
                # (all_flattened_tails_with_root) -> label
            )
            lab_loss = (
                F.cross_entropy(
                    flat_lab_logits,
                    flat_lab_targets,
                    reduction="sum",
                )
                / batch["stats"]["dep_words"]
            )
            return {
                "arc_loss": arc_loss,
                "lab_loss": lab_loss,
                "loss": arc_loss + lab_loss,
            }
        else:
            return {
                "arc_logits": arc_logits,
                "arc_labels": lab_logits.argmax(-1),
            }

    def postprocess(
        self,
        docs: Sequence[Doc],
        results: BatchOutput,
        inputs: List[Dict[str, Any]],
    ) -> Sequence[Doc]:
        # Preprocessed docs should still be in the cache
        # (context, head=words + 1, tail=words + 1), ie head -> tail
        contexts = [ctx for sample in inputs for ctx in sample["$contexts"]]

        for ctx, arc_logits, arc_labels in zip(
            contexts,
            results["arc_logits"].detach().cpu().numpy(),
            results["arc_labels"].detach().cpu(),
        ):
            ctx: Span
            if self.decoding_mode == "greedy":
                tail_to_head_idx = arc_logits.argmax(-1)
            else:
                tail_to_head_idx = chuliu_edmonds_one_root(arc_logits)
                tail_to_head_idx = torch.as_tensor(tail_to_head_idx)

            # lab_logits: (tail=words+1, head=words+1, labels) -> prob
            # arc_logits: (tail=words+1) -> head_idx
            labels = arc_labels[torch.arange(arc_labels.shape[0]), tail_to_head_idx]

            # Set arc and dep rel on the Span context
            for tail_idx, (head_idx, label) in enumerate(
                zip(tail_to_head_idx.tolist(), labels.tolist())
            ):
                if head_idx == 0:
                    continue
                head = ctx[head_idx - 1]
                tail = ctx[tail_idx - 1]
                tail.head = head
                tail.dep_ = self.labels[label]

        return docs



from __future__ import annotations

import warnings
from typing import Any, Dict, Optional, Sequence

import foldedtensor as ft
import torch
from confit import VisibleDeprecationWarning
from spacy.tokens import Doc, Span
from typing_extensions import Literal, TypedDict

from edsnlp.core.pipeline import Pipeline
from edsnlp.core.torch_component import BatchInput
from edsnlp.pipes.base import BaseComponent
from edsnlp.pipes.trainable.embeddings.typing import (
    SpanEmbeddingComponent,
    WordEmbeddingComponent,
)
from edsnlp.utils.filter import align_spans

SpanPoolerBatchInput = TypedDict(
    "SpanPoolerBatchInput",
    {
        "embedding": BatchInput,
        "begins": ft.FoldedTensor,
        "ends": ft.FoldedTensor,
        "sequence_idx": torch.Tensor,
        "stats": TypedDict("SpanPoolerBatchStats", {"spans": int}),
    },
)
"""
embeds: torch.FloatTensor
    Token embeddings to predict the tags from
begins: torch.LongTensor
    Begin offsets of the spans
ends: torch.LongTensor
    End offsets of the spans
sequence_idx: torch.LongTensor
    Sequence (cf Embedding spans) index of the spans
"""

SpanPoolerBatchOutput = TypedDict(
    "SpanPoolerBatchOutput",
    {
        "embeddings": ft.FoldedTensor,
    },
)


class SpanPooler(SpanEmbeddingComponent, BaseComponent):
    """
    The `eds.span_pooler` component is a trainable span embedding component. It
    generates span embeddings from a word embedding component and a span getter. It can
    be used to train a span classifier, as in `eds.span_classifier`.

    Parameters
    ----------
    nlp: PipelineProtocol
        The pipeline object
    name: str
        Name of the component
    embedding : WordEmbeddingComponent
        The word embedding component
    pooling_mode: Literal["max", "sum", "mean"]
        How word embeddings are aggregated into a single embedding per span.
    hidden_size : Optional[int]
        The size of the hidden layer. If None, no projection is done and the output
        of the span pooler is used directly.
    """

    def __init__(
        self,
        nlp: Optional[Pipeline] = None,
        name: str = "span_pooler",
        *,
        embedding: WordEmbeddingComponent,
        pooling_mode: Literal["max", "sum", "mean"] = "mean",
        hidden_size: Optional[int] = None,
        span_getter: Any = None,
    ):
        if span_getter is not None:
            warnings.warn(
                "The `span_getter` parameter of the `eds.span_pooler` component is "
                "deprecated. Please use the `span_getter` parameter of the "
                "`eds.span_classifier` or `eds.span_linker` components instead.",
                VisibleDeprecationWarning,
            )
        sub_span_getter = getattr(embedding, "span_getter", None)
        if sub_span_getter is not None and span_getter is None:  # pragma: no cover
            self.span_getter = sub_span_getter
        sub_context_getter = getattr(embedding, "context_getter", None)
        if sub_context_getter is not None:  # pragma: no cover
            self.context_getter = sub_context_getter

        self.output_size = embedding.output_size if hidden_size is None else hidden_size

        super().__init__(nlp, name)

        self.pooling_mode = pooling_mode
        self.span_getter = span_getter
        self.embedding = embedding
        self.projector = (
            torch.nn.Linear(self.embedding.output_size, hidden_size)
            if hidden_size is not None
            else torch.nn.Identity()
        )

    def feed_forward(self, span_embeds: torch.Tensor) -> torch.Tensor:
        return self.projector(span_embeds)

    def preprocess(
        self,
        doc: Doc,
        *,
        spans: Optional[Sequence[Span]] = None,
        contexts: Optional[Sequence[Span]] = None,
        pre_aligned: bool = False,
        **kwargs,
    ) -> Dict[str, Any]:
        contexts = contexts if contexts is not None else [doc[:]]

        sequence_idx = []
        begins = []
        ends = []

        contexts_to_idx = {span: i for i, span in enumerate(contexts)}
        assert not pre_aligned or len(spans) == len(contexts), (
            "When `pre_aligned` is True, the number of spans and contexts must be the "
            "same."
        )
        aligned_contexts = (
            [[c] for c in contexts]
            if pre_aligned
            else align_spans(contexts, spans, sort_by_overlap=True)
        )
        for i, (span, ctx) in enumerate(zip(spans, aligned_contexts)):
            if len(ctx) == 0 or ctx[0].start > span.start or ctx[0].end < span.end:
                raise Exception(
                    f"Span {span.text!r} is not included in at least one embedding "
                    f"span: {[s.text for s in ctx]}"
                )
            start = ctx[0].start
            sequence_idx.append(contexts_to_idx[ctx[0]])
            begins.append(span.start - start)
            ends.append(span.end - start)
        return {
            "begins": begins,
            "ends": ends,
            "sequence_idx": sequence_idx,
            "num_sequences": len(contexts),
            "embedding": self.embedding.preprocess(doc, contexts=contexts, **kwargs),
            "stats": {"spans": len(begins)},
        }

    def collate(self, batch: Dict[str, Sequence[Any]]) -> SpanPoolerBatchInput:
        sequence_idx = []
        offset = 0
        for indices, seq_length in zip(batch["sequence_idx"], batch["num_sequences"]):
            sequence_idx.extend([offset + idx for idx in indices])
            offset += seq_length

        collated: SpanPoolerBatchInput = {
            "embedding": self.embedding.collate(batch["embedding"]),
            "begins": ft.as_folded_tensor(
                batch["begins"],
                data_dims=("span",),
                full_names=("sample", "span"),
                dtype=torch.long,
            ),
            "ends": ft.as_folded_tensor(
                batch["ends"],
                data_dims=("span",),
                full_names=("sample", "span"),
                dtype=torch.long,
            ),
            "sequence_idx": torch.as_tensor(sequence_idx),
            "stats": {"spans": sum(batch["stats"]["spans"])},
        }
        return collated

    # noinspection SpellCheckingInspection
    def forward(self, batch: SpanPoolerBatchInput) -> SpanPoolerBatchOutput:
        """
        Apply the span classifier module to the document embeddings and given spans to:
        - compute the loss
        - and/or predict the labels of spans
        If labels are predicted, they are assigned to the `additional_outputs`
        dictionary.

        Parameters
        ----------
        batch: SpanPoolerBatchInput
            The input batch

        Returns
        -------
        BatchOutput
        """
        device = next(self.parameters()).device
        if len(batch["begins"]) == 0:
            span_embeds = torch.empty(0, self.output_size, device=device)
            return {
                "embeddings": batch["begins"].with_data(span_embeds),
            }

        embeds = self.embedding(batch["embedding"])["embeddings"]
        _, n_words, dim = embeds.shape
        device = embeds.device

        flat_begins = n_words * batch["sequence_idx"] + batch["begins"].as_tensor()
        flat_ends = n_words * batch["sequence_idx"] + batch["ends"].as_tensor()
        flat_embeds = embeds.view(-1, dim)
        flat_indices = torch.cat(
            [
                torch.arange(b, e, device=device)
                for b, e in zip(flat_begins.cpu().tolist(), flat_ends.cpu().tolist())
            ]
        ).to(device)
        offsets = (flat_ends - flat_begins).cumsum(0).roll(1)
        offsets[0] = 0
        span_embeds = torch.nn.functional.embedding_bag(  # type: ignore
            input=flat_indices,
            weight=flat_embeds,
            offsets=offsets,
            mode=self.pooling_mode,
        )
        span_embeds = self.feed_forward(span_embeds)

        return {
            "embeddings": batch["begins"].with_data(span_embeds),
        }



from __future__ import annotations

import pickle
import sys
from typing import Optional, Union

import dill

from edsnlp.core.stream import Stream
from edsnlp.data.base import BaseWriter, BatchWriter
from edsnlp.data.spark import SparkReader, SparkWriter
from edsnlp.utils.collections import flatten, flatten_once
from edsnlp.utils.stream_sentinels import StreamSentinel

try:
    from koalas.dataframe import DataFrame as KoalasDataFrame
except ImportError:  # pragma: no cover
    KoalasDataFrame = None


class Broadcasted:
    def __init__(self, value):
        self.value = value

    def __getstate__(self):
        return dill.dumps(self.value, byref=True, recurse=True)

    def __setstate__(self, state):  # pragma: no cover
        self.value = dill.loads(state)


def execute_spark_backend(
    stream: Stream,
):
    """
    This execution mode uses Spark to parallelize the processing of the documents.
    The documents are first stored in a Spark DataFrame (if it was not already the case)
    and then processed in parallel using Spark.

    Beware, if the original reader was not a SparkReader (`edsnlp.data.from_spark`), the
    *local docs* → *spark dataframe* conversion might take some time, and the whole
    process might be slower than using the `multiprocessing` backend.
    """
    import pyspark.sql.types as T
    from pyspark.sql import SparkSession

    from edsnlp.utils.spark_dtypes import schema_warning, spark_interpret_dicts_as_rows

    try:
        getActiveSession = SparkSession.getActiveSession
    except AttributeError:

        def getActiveSession() -> Optional["SparkSession"]:  # pragma: no cover
            from pyspark import SparkContext

            sc = SparkContext._active_spark_context
            if sc is None:
                return None
            else:
                assert sc._jvm is not None
                if sc._jvm.SparkSession.getActiveSession().isDefined():
                    SparkSession(sc, sc._jvm.SparkSession.getActiveSession().get())
                    try:
                        return SparkSession._activeSession
                    except AttributeError:
                        try:
                            return SparkSession._instantiatedSession
                        except AttributeError:
                            return None
                else:
                    return None

    # Get current spark session
    spark = getActiveSession() or SparkSession.builder.getOrCreate()

    reader = stream.reader
    writer: Union[BaseWriter, BatchWriter] = stream.writer

    if isinstance(reader, SparkReader):
        df = reader.data
        assert not reader.loop, "Looping is not supported with Spark backend."
        df = (
            df.sample(fraction=1.0, seed=reader.rng.getrandbits(32))
            if reader.shuffle
            else df
        )
    else:
        with spark_interpret_dicts_as_rows():
            df = spark.sparkContext.parallelize(
                [
                    {"content": pickle.dumps(item, -1)}
                    for item in reader.read_records()
                    if not isinstance(item, StreamSentinel)
                ]
            ).toDF(T.StructType([T.StructField("content", T.BinaryType())]))

    def make_torch_pipe(torch_pipe, disable_after):  # pragma: no cover
        def wrapped(batches):
            for batch in batches:
                batch_id = hash(tuple(id(x) for x in batch))
                torch_pipe.enable_cache(batch_id)
                batch = torch_pipe.batch_process(batch)
                if disable_after:
                    torch_pipe.disable_cache(batch_id)
                yield batch

        return wrapped

    def process_partition(items):  # pragma: no cover
        stream: Stream = bc.value
        writer = stream.writer
        try:
            sys.modules["torch"].set_grad_enabled(False)
        except (AttributeError, KeyError):
            pass
        stages = stream._make_stages(split_torch_pipes=True)

        if not isinstance(stream.reader, SparkReader):
            items = (pickle.loads(row.content) for row in items)
        else:
            items = (item.asDict(recursive=True) for item in items)

        items = (task for item in items for task in stream.reader.extract_task(item))

        for stage_idx, stage in enumerate(stages):
            for op in stage.cpu_ops:
                items = op(items)

            if stage.gpu_op is not None:
                pipe = make_torch_pipe(stage.gpu_op, stage_idx == len(stages) - 2)
                items = pipe(items)

        if writer is not None:
            items = (writer.handle_record(item) for item in items)

        results = []
        if getattr(writer, "write_in_worker", None) is True:
            items = writer.batch_fn(items, writer.batch_size)
            # get the 1st element (2nd is the count)
            for item in items:
                item, count = writer.handle_batch(item)
                results.append(item)
        else:
            results = list(items)

        if isinstance(writer, SparkWriter):
            return list(flatten(results))
        else:
            return [{"content": pickle.dumps(results, -1)}]

    with stream.eval():
        bc = Broadcasted(stream.worker_copy())

        if isinstance(writer, SparkWriter):
            rdd = df.rdd.mapPartitions(process_partition)
            with spark_interpret_dicts_as_rows():
                results = spark.createDataFrame(rdd, schema=writer.dtypes)

            if writer.dtypes is None and writer.show_dtypes:
                schema_warning(results.schema)
            return results

        items = flatten_once(
            pickle.loads(item["content"])
            for item in df.rdd.mapPartitions(process_partition).toLocalIterator()
        )

        if getattr(writer, "write_in_worker", None) is False:
            writer: BatchWriter
            items = writer.batch_fn(items, writer.batch_size)
            # get the 1st element (2nd is the count)
            items = (writer.handle_batch(b)[0] for b in items)
        return items if writer is None else writer.consolidate(items)



import os
import subprocess
import sys
from io import BytesIO

import pytest
from confit import Config
from confit.errors import ConfitValidationError
from confit.registry import validate_arguments
from spacy.tokens import Doc

import edsnlp
import edsnlp.pipes as eds
from edsnlp import Pipeline, registry
from edsnlp.core.registries import DraftPipe
from edsnlp.pipes.base import BaseComponent

try:
    import torch.nn
except ImportError:
    torch = None


class CustomClass:
    pass

    def __call__(self, doc: Doc) -> Doc:
        return doc


def test_add_pipe_factory():
    model = edsnlp.blank("eds")
    model.add_pipe("eds.normalizer", name="normalizer")
    assert "normalizer" in model.pipe_names
    assert model.has_pipe("normalizer")

    model.add_pipe("eds.sentences", name="sentences")
    assert "sentences" in model.pipe_names
    assert model.has_pipe("sentences")

    with pytest.raises(ValueError):
        model.get_pipe("missing-pipe")


def test_add_pipe_component():
    model = edsnlp.blank("eds")
    model.add_pipe(eds.normalizer(nlp=model), name="normalizer")
    assert "normalizer" in model.pipe_names
    assert model.has_pipe("normalizer")
    assert model.pipes.normalizer is model.get_pipe("normalizer")

    model.add_pipe(eds.sentences(nlp=model), name="sentences")
    assert "sentences" in model.pipe_names
    assert model.has_pipe("sentences")
    assert model.pipes.sentences is model.get_pipe("sentences")

    with pytest.raises(ValueError):
        model.add_pipe(
            eds.sentences(nlp=model, name="sentences"),
            config={"punct_chars": ".?!"},
        )

    with pytest.raises(ValueError):
        model.add_pipe(CustomClass())


def test_sequence(frozen_ml_nlp: Pipeline):
    assert len(frozen_ml_nlp.pipeline) == 3
    assert list(frozen_ml_nlp.pipeline) == [
        ("sentences", frozen_ml_nlp.get_pipe("sentences")),
        ("transformer", frozen_ml_nlp.get_pipe("transformer")),
        ("ner", frozen_ml_nlp.get_pipe("ner")),
    ]
    assert list(frozen_ml_nlp.torch_components()) == [
        ("transformer", frozen_ml_nlp.get_pipe("transformer")),
        ("ner", frozen_ml_nlp.get_pipe("ner")),
    ]


def test_disk_serialization(tmp_path, ml_nlp):
    nlp = ml_nlp

    assert nlp.get_pipe("transformer").stride == 96
    ner = nlp.get_pipe("ner")
    ner.update_labels(["PERSON", "GIFT"])

    os.makedirs(tmp_path / "model", exist_ok=True)
    # by default, vocab is excluded
    nlp.to_disk(tmp_path / "model", exclude=set())

    assert (tmp_path / "model" / "config.cfg").exists()
    assert (tmp_path / "model" / "ner" / "parameters.safetensors").exists()
    assert (tmp_path / "model" / "transformer" / "parameters.safetensors").exists()
    # fmt: off
    assert (
            (tmp_path / "model" / "transformer" / "pytorch_model.bin").exists() or
            (tmp_path / "model" / "transformer" / "model.safetensors").exists()
    )
    # fmt: on

    assert (tmp_path / "model" / "config.cfg").read_text() == (
        config_str.replace("components = ${components}\n", "").replace(
            "prajjwal1/bert-tiny", "./transformer"
        )
    )

    nlp = edsnlp.load(
        tmp_path / "model",
        overrides={"components": {"transformer": {"stride": 64}}},
    )
    assert nlp.get_pipe("ner").labels == ["PERSON", "GIFT"]
    assert nlp.get_pipe("transformer").stride == 64


config_str = """\
[nlp]
lang = "eds"
pipeline = ["sentences", "transformer", "ner"]
components = ${components}

[nlp.tokenizer]
@tokenizers = "eds.tokenizer"

[components]

[components.sentences]
@factory = "eds.sentences"

[components.transformer]
@factory = "eds.transformer"
model = "prajjwal1/bert-tiny"
window = 128
stride = 96

[components.ner]
@factory = "eds.ner_crf"
embedding = ${components.transformer}
mode = "independent"
target_span_getter = ["ents", "ner-preds"]
labels = ["PERSON", "GIFT"]
infer_span_setter = false
window = 40
stride = 20

[components.ner.span_setter]
ents = true

"""


@pytest.mark.skipif(torch is None, reason="torch not installed")
def test_validate_config():
    @validate_arguments
    def function(model: Pipeline):
        assert len(model.pipe_names) == 3

    function(Config.from_str(config_str).resolve(registry=registry)["nlp"])


def test_torch_module(frozen_ml_nlp: Pipeline):
    with frozen_ml_nlp.train(True):
        for name, component in frozen_ml_nlp.torch_components():
            assert component.training is True

    with frozen_ml_nlp.train(False):
        for name, component in frozen_ml_nlp.torch_components():
            assert component.training is False

    frozen_ml_nlp.to("cpu")


def test_cache(frozen_ml_nlp: Pipeline):
    from edsnlp.core.torch_component import _caches

    text = "Ceci est un exemple"
    frozen_ml_nlp(text)

    doc = frozen_ml_nlp.make_doc(text)
    with frozen_ml_nlp.cache():
        for name, pipe in frozen_ml_nlp.pipeline:
            # This is a hack to get around the ambiguity
            # between the __call__ method of Pytorch modules
            # and the __call__ methods of spacy components
            if hasattr(pipe, "batch_process"):
                doc = next(iter(pipe.batch_process([doc])))
            else:
                doc = pipe(doc)
        trf_forward_cache_entries = [
            key
            for key in _caches["default"]
            if isinstance(key, tuple) and key[0] == "forward"
        ]
        assert len(trf_forward_cache_entries) == 2

    assert len(_caches) == 0


def test_select_pipes(frozen_ml_nlp: Pipeline):
    text = "Ceci est un exemple"
    with frozen_ml_nlp.select_pipes(enable=["transformer", "ner"]):
        assert len(frozen_ml_nlp.disabled) == 1
        assert not frozen_ml_nlp(text).has_annotation("SENT_START")
    assert len(frozen_ml_nlp.disabled) == 0


@pytest.mark.skip(reason="Deprecated behavior")
def test_different_names():
    nlp = edsnlp.blank("eds")

    extractor = eds.sentences(nlp=nlp, name="custom_name")

    with pytest.raises(ValueError) as exc_info:
        nlp.add_pipe(extractor, name="sentences")

    assert (
        "The provided name 'sentences' does not "
        "match the name of the component 'custom_name'."
    ) in str(exc_info.value)


@pytest.mark.skipif(torch is None, reason="torch not installed")
def test_load_config(run_in_test_dir):
    nlp = edsnlp.load("training/qlf_config.yml")
    assert nlp.pipe_names == [
        "normalizer",
        "sentencizer",
        "covid",
        "qualifier",
    ]


fail_config = """
[nlp]
lang = "eds"
pipeline = ["transformer", "ner"]

[nlp.tokenizer]
@tokenizers = "eds.tokenizer"

[components]

[components.transformer]
@factory = "eds.transformer"
model = "prajjwal1/bert-tiny"
window = 128
stride = 96

[components.ner]
@factory = "eds.ner_crf"
embedding = ${components.transformer}
mode = "error-mode"
span_setter = "ents"
"""


@pytest.mark.skipif(torch is None, reason="torch not installed")
def test_config_validation_error():
    with pytest.raises(ConfitValidationError) as e:
        Pipeline.from_config(Config.from_str(fail_config))

    assert "1 validation error for" in str(e.value)
    assert "got 'error-mode'" in str(e.value)


@edsnlp.registry.factory.register("test_wrapper", spacy_compatible=False)
class WrapperComponent:
    def __init__(self, *, copy_list, copy_dict, sub):
        pass


fail_config_sub = """
nlp:
    lang: "eds"
    components:
        wrapper:
            "@factory": "test_wrapper"

            copy_list:
                - ${nlp.components.wrapper.sub}

            copy_dict:
                key: ${nlp.components.wrapper.sub}

            sub:
                "@factory": "eds.matcher"
                terms: 100.0  # clearly wrong

        matcher_copy: ${nlp.components.wrapper.sub}
"""


def test_config_sub_validation_error():
    with pytest.raises(ConfitValidationError):
        Pipeline.from_config(Config.from_yaml_str(fail_config_sub))

    fix = {"nlp": {"components": {"wrapper": {"sub": {"terms": {"pattern": ["ok"]}}}}}}
    Pipeline.from_config(Config.from_yaml_str(fail_config_sub).merge(fix))


def test_add_pipe_validation_error():
    model = edsnlp.blank("eds")
    with pytest.raises(ConfitValidationError) as e:
        model.add_pipe("eds.covid", name="extractor", config={"foo": "bar"})

    assert str(e.value) == (
        "1 validation error for "
        "edsnlp.pipes.ner.covid.factory.create_component()\n"
        "-> extractor.foo\n"
        "   unexpected keyword argument"
    )


def test_spacy_component():
    nlp = edsnlp.blank("fr")
    nlp.add_pipe("sentencizer")


def test_rule_based_pipeline():
    nlp = edsnlp.blank("eds")
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe("eds.covid")

    assert nlp.pipe_names == ["normalizer", "covid"]
    assert nlp.get_pipe("normalizer") == nlp.pipeline[0][1]
    assert nlp.has_pipe("covid")

    with pytest.raises(ValueError) as exc_info:
        nlp.get_pipe("unknown")

    assert str(exc_info.value) == "Pipe 'unknown' not found in pipeline."

    doc = nlp.make_doc("Mon patient a le covid")

    new_doc = nlp(doc)

    assert len(doc.ents) == 1
    assert new_doc is doc

    assert nlp.get_pipe_meta("covid").assigns == ["doc.ents", "doc.spans"]


def test_torch_save(ml_nlp):
    import torch

    ml_nlp.get_pipe("ner").update_labels(["LOC", "PER"])
    buffer = BytesIO()
    torch.save(ml_nlp, buffer)
    buffer.seek(0)
    nlp = torch.load(buffer, weights_only=False)
    assert nlp.get_pipe("ner").labels == ["LOC", "PER"]
    assert len(list(nlp("Une phrase. Deux phrases.").sents)) == 2


def test_parameters(frozen_ml_nlp):
    assert len(list(frozen_ml_nlp.parameters())) == 42


def test_missing_factory(nlp):
    with pytest.raises(ValueError) as exc_info:
        nlp.add_pipe("__test_missing_pipe__")

    assert "__test_missing_pipe__" in str(exc_info.value)


@edsnlp.registry.factory("custom-curry-test")
class CustomComponent(BaseComponent):
    def __init__(self, nlp, name):
        self.nlp = nlp

    def __call__(self, doc):
        return doc


def test_curried_nlp_pipe():
    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences(name="my-sentences"))
    nlp.add_pipe(eds.normalizer())
    nlp.add_pipe(eds.sections(), name="sections")
    pipe = CustomComponent()

    assert isinstance(pipe, DraftPipe)
    err = (
        f"This Draft[{pipe._func.__qualname__}] component has not been instantiated "
        f"yet, likely because it was missing an `nlp` pipeline argument. You should "
        f"either:\n"
        "- add it to a pipeline: `pipe = nlp.add_pipe(pipe)`\n"
        "- or fill its `nlp` argument: `pipe = factory(nlp=nlp, ...)`"
    )
    with pytest.raises(TypeError) as exc_info:
        pipe("Demo texte")
    assert str(exc_info.value) == err

    with pytest.raises(TypeError) as exc_info:
        pipe.forward("Demo texte")
    assert str(exc_info.value) == err

    nlp.add_pipe(pipe, name="custom")

    assert nlp.pipes.custom.nlp is nlp

    assert nlp.pipe_names == ["my-sentences", "normalizer", "sections", "custom"]


@pytest.mark.skipif(
    sys.version_info < (3, 8),
    reason="Can't run on GH CI with Python 3.7",
)
@pytest.mark.skipif(torch is None, reason="torch not installed")
def test_huggingface():
    nlp = edsnlp.load(
        "AP-HP/dummy-ner",
        auto_update=True,
        install_dependencies=True,
    )
    doc = nlp("On lui prescrit du paracetamol à 500mg.")
    assert doc.ents[0].text == "paracetamol"
    assert doc.ents[1].text == "500mg"

    # Try loading it twice for coverage
    edsnlp.load(
        "AP-HP/dummy-ner",
        auto_update=True,
        install_dependencies=True,
    )

    subprocess.run(["pip", "uninstall", "dummy-pip-package", "-y"], check=True)


@pytest.mark.skipif(
    sys.version_info < (3, 8),
    reason="Can't run on GH CI with Python 3.7",
)
def test_missing_huggingface():
    with pytest.raises(ValueError) as exc_info:
        edsnlp.load(
            "AP-HP/does-not-exist",
            auto_update=True,
        )

    assert "The load function expects either :" in str(exc_info.value)


def test_repr(frozen_ml_nlp):
    with frozen_ml_nlp.select_pipes(disable=["sentences"]):
        assert (
            repr(frozen_ml_nlp)
            == """\
Pipeline(lang=eds, pipes={
  "sentences": [disabled] eds.sentences,
  "transformer": eds.transformer,
  "ner": eds.ner_crf
})"""
        )


@edsnlp.registry.factory.register("test_nlp_less", spacy_compatible=False)
class NlpLessComponent:
    def __init__(self, nlp=None, name: str = "nlp_less", *, value: int):
        self.value = value
        self.name = name

    def __call__(self, doc):
        return doc


def test_nlp_less_component():
    component = NlpLessComponent(value=42)
    assert component.value == 42

    config = """
[component]
@factory = "test_nlp_less"
value = 42
"""
    component = Config.from_str(config).resolve(registry=registry)["component"]
    assert component.value == 42



import abc
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np
from pydantic import NonNegativeInt
from spacy.tokens import Doc, Span

from edsnlp import registry
from edsnlp.utils.filter import filter_spans
from edsnlp.utils.typing import AsList, Validated

SeqStr = AsList[str]
SpanFilter = Union[bool, SeqStr]

SpanSetterMapping = Dict[str, SpanFilter]
SpanGetterMapping = Dict[str, SpanFilter]

SpanGetter = Union[
    SpanGetterMapping,
    Callable[[Doc], Iterable[Span]],
]
SpanSetter = Union[
    SpanSetterMapping,
    Callable[[Doc, Iterable[Span]], Any],
]


def get_spans(doclike, span_getter, deduplicate=True):
    if span_getter is None:
        yield doclike[:]
        return
    if callable(span_getter):
        yield from span_getter(doclike)
        return
    seen = set()
    for k, span_filter in span_getter.items():
        doc = doclike if isinstance(doclike, Doc) else doclike.doc
        if k == "*":
            candidates = (s for grp in doc.spans.values() for s in grp)
        elif k == "ents":
            candidates = doc.ents
        elif k == "sents":
            candidates = doc.sents
        else:
            candidates = doc.spans.get(k, ())
        if not isinstance(doclike, Doc):
            candidates = (
                s
                for s in candidates
                if not (s.end < doclike.start or s.start > doclike.end)
            )
        for span in candidates:
            if (span_filter is True) or (span.label_ in span_filter):
                if span not in seen:
                    yield span
                    if deduplicate:
                        seen.add(span)


def get_spans_with_group(doclike, span_getter):
    if span_getter is None:
        yield doclike[:], None
        return
    if callable(span_getter):
        yield from span_getter(doclike)
        return
    for key, span_filter in span_getter.items():
        doc = doclike if isinstance(doclike, Doc) else doclike.doc
        if key == "*":
            candidates = (
                (span, name) for name, group in doc.spans.items() for span in group
            )
        elif key == "sents":
            candidates = ((sent, key) for sent in doc.sents)
        elif key == "ents":
            candidates = ((ent, key) for ent in doc.ents)
        else:
            candidates = ((span, key) for span in doc.spans.get(key, ()))
        if not isinstance(doclike, Doc):
            candidates = (
                x
                for x in candidates
                if not (x[0].end < doclike.start or x[0].start > doclike.end)
            )
        if span_filter is True:
            yield from candidates
        else:
            for span, group in candidates:
                if span.label_ in span_filter:
                    yield span, group


def set_spans(doc, matches, span_setter):
    # TODO: prevent "sents" in span_getter ?
    if callable(span_setter):
        span_setter(doc, matches)
    else:
        match_all = []
        label_to_group = defaultdict(list)
        for name, spans_filter in span_setter.items():
            if name != "ents" and name != "*":
                doc.spans.setdefault(name, [])
            if spans_filter:
                if spans_filter is True:
                    match_all.append(name)
                else:
                    for label in spans_filter:
                        label_to_group[label].append(name)

        new_ents = [] if "ents" in span_setter else None

        for span in matches:
            for group in match_all + label_to_group[span.label_]:
                if group == "ents":
                    new_ents.append(span)
                elif group == "*":
                    doc.spans.setdefault(span.label_, []).append(span)
                else:
                    doc.spans[group].append(span)
        if new_ents is not None:
            doc.ents = filter_spans((*new_ents, *doc.ents))
    return doc


def validate_span_setter(value: Union[SeqStr, Dict[str, SpanFilter]]) -> SpanSetter:
    if callable(value):
        return value
    if isinstance(value, str):
        return {value: True}
    if isinstance(value, list):
        return {group: True for group in value}
    elif isinstance(value, dict):
        new_value = {}
        for k, v in value.items():
            if isinstance(v, bool):
                new_value[k] = v
            elif isinstance(v, str):
                new_value[k] = [v]
            elif isinstance(v, list) and all(isinstance(i, str) for i in v):
                new_value[k] = v
            else:
                raise TypeError(
                    f"Invalid entry {value} ({type(value)}) for SpanSetterArg, "
                    f"expected bool/string(s), dict of bool/string(s) or callable"
                )
        return new_value
    else:
        raise TypeError(
            f"Invalid entry {value} ({type(value)}) for SpanSetterArg, "
            f"expected bool/string(s), dict of bool/string(s) or callable"
        )


def validate_span_getter(
    value: Union[SeqStr, Dict[str, SpanFilter]], optional: bool = False
) -> SpanSetter:
    if value is None:
        if optional:
            return None
        raise ValueError(
            "Mising entry for SpanGetterArg, "
            "expected bool/string(s), dict of bool/string(s) or callable"
        )
    if callable(value):
        return value
    if isinstance(value, str):
        return {value: True}
    if isinstance(value, list):
        return {group: True for group in value}
    elif isinstance(value, dict):
        new_value = {}
        for k, v in value.items():
            if isinstance(v, bool):
                new_value[k] = v
            elif isinstance(v, str):
                new_value[k] = [v]
            elif isinstance(v, list) and all(isinstance(i, str) for i in v):
                new_value[k] = v
            else:
                raise TypeError(
                    f"Invalid entry {value} ({type(value)}) for SpanGetterArg, "
                    f"expected bool/string(s), dict of bool/string(s) or callable"
                )
        return new_value
    else:
        raise TypeError(
            f"Invalid entry {value} ({type(value)}) for SpanGetterArg, "
            f"expected bool/string(s), dict of bool/string(s) or callable"
        )


class SpanSetterArg(Validated):
    """
    Valid values for the `span_setter` argument of a component can be :

    - a (doc, matches) -> None callable
    - a span group name
    - a list of span group names
    - a dict of group name to True or list of labels

    The group name `"ents"` is a special case, and will add the matches to `doc.ents`

    Examples
    --------
    - `span_setter=["ents", "ckd"]` will add the matches to both `doc.ents` and
    `doc.spans["ckd"]`. It is equivalent to `{"ents": True, "ckd": True}`.
    - `span_setter={"ents": ["foo", "bar"]}` will add the matches with label
    "foo" and "bar" to `doc.ents`.
    - `span_setter="ents"` will add all matches only to `doc.ents`.
    - `span_setter="ckd"` will add all matches only to `doc.spans["ckd"]`.
    """

    @classmethod
    def validate(cls, value, config=None) -> SpanSetter:
        return validate_span_setter(value)


class SpanGetterArg(Validated):
    """
    Valid values for the `span_getter` argument of a component can be :

    - a (doc) -> spans callable
    - a span group name
    - a list of span group names
    - a dict of group name to True or list of labels

    The group name `"ents"` is a special case, and will get the matches from `doc.ents`

    Examples
    --------
    - `span_getter=["ents", "ckd"]` will get the matches from both `doc.ents` and
    `doc.spans["ckd"]`. It is equivalent to `{"ents": True, "ckd": True}`.
    - `span_getter={"ents": ["foo", "bar"]}` will get the matches with label
    "foo" and "bar" from `doc.ents`.
    - `span_getter="ents"` will get all matches from `doc.ents`.
    - `span_getter="ckd"` will get all matches from `doc.spans["ckd"]`.
    """

    @classmethod
    def validate(cls, value, config=None) -> SpanSetter:
        return validate_span_getter(value)


if TYPE_CHECKING:
    SpanGetterArg = Union[  # noqa: F811
        str,
        Sequence[str],
        SpanGetterMapping,
        Callable[[Doc], Iterable[Span]],
    ]
    SpanSetterArg = Union[  # noqa: F811
        str,
        Sequence[str],
        SpanSetterMapping,
        Callable[[Doc, Iterable[Span]], Any],
    ]


@registry.misc.register("eds.span_context_getter")
class make_span_context_getter:
    """
    Create a span context getter.

    Parameters
    ----------
    context_words : Union[NonNegativeInt, Tuple[NonNegativeInt, NonNegativeInt]]
        Minimum number of words to include on each side of the span. It could be
        asymmetric. For example (5,2) will include 5 words before the start of the
        span and 2 after the end of the span
    context_sents : Optional[
            Union[NonNegativeInt, Tuple[NonNegativeInt, NonNegativeInt]]
        ] = 1
        Minimum number of sentences to include on each side of the span:

        - 0: don't use sentences to build the context.
        - 1: include the sentence of the span.
        - n: include n-1 sentences on each side of the span + the sentence of the span


        By default, 0 if the document has no sentence annotations, 1 otherwise.
    """

    def __init__(
        self,
        context_words: Union[NonNegativeInt, Tuple[NonNegativeInt, NonNegativeInt]],
        context_sents: Union[NonNegativeInt, Tuple[NonNegativeInt, NonNegativeInt]],
        span_getter: Optional[SpanGetterArg] = None,
    ):
        if isinstance(context_words, int):
            self.n_words_left, self.n_words_right = (context_words, context_words)
        else:
            self.n_words_left, self.n_words_right = context_words

        if isinstance(context_sents, int):
            self.context_sents_left, self.context_sents_right = (
                context_sents,
                context_sents,
            )
        else:
            self.context_sents_left, self.context_sents_right = context_sents
            assert sum(context_sents) != 1, (
                "Asymmetric sentence context should not be (0,1) or (1,0)"
            )
        self.span_getter = validate_span_getter(span_getter, optional=True)

    def __call__(self, span: Union[Doc, Span]) -> Union[Span, List[Span]]:
        if isinstance(span, Doc):  # pragma: no cover
            return [self(s) for s in get_spans(span, self.span_getter)]

        n_sents_left, n_sents_right = self.context_sents_left, self.context_sents_right
        n_words_left = self.n_words_left
        n_words_right = self.n_words_right

        start = max(0, span.start - n_words_left)
        end = min(len(span.doc), span.end + n_words_right)

        n_sents_max = max(n_sents_left, n_sents_right)
        if n_sents_max > 0:
            min_start_sent = start
            max_end_sent = end
            if n_sents_left == 1:
                sent = span.sent
                min_start_sent = sent.start
            if n_sents_right == 1:
                sent = span.sent
                max_end_sent = sent.end
            if (n_sents_left != 1) or (n_sents_right != 1):
                sents = list(span.doc.sents) if n_sents_max > 1 else []
                sent_i = sents.index(span.sent)
                min_start_sent = sents[max(0, sent_i - n_sents_left + 1)].start
                max_end_sent = sents[
                    min(len(sents) - 1, sent_i + n_sents_right - 1)
                ].end
            start = min(start, min_start_sent)
            end = max(end, max_end_sent)

        return span.doc[start:end]


class ContextWindowMeta(abc.ABCMeta):
    pass


class ContextWindow(Validated, abc.ABC, metaclass=ContextWindowMeta):
    """
    A ContextWindow specifies how much additional context (such as sentences or words)
    should be included relative to an anchor span. For example, one might define a
    context window that extracts the sentence immediately preceding and following the
    anchor span, or one that extends the span by a given number of words before and
    after.

    ContextWindow objects can be combined using logical operations to create more
    complex context windows. For example, one can create a context window that includes
    either words from a -10 to +10 range or words from the sentence.


    Examples
    --------

    ```python
    from confit import validate_arguments
    from spacy.tokens import Span

    import edsnlp
    from edsnlp.utils.span_getters import ContextWindow


    @validate_arguments
    def apply_context(span: Span, ctx: ContextWindow):
        # ctx will be parsed and cast as a ContextWindow
        return ctx(span)


    nlp = edsnlp.blank("eds")
    nlp.add_pipe("eds.sentences")

    doc = nlp("A first sentence. A second sentence, longer this time. A third.")
    span = doc[5:6]  # "second"

    # Will return a span with the 10 words before and after the span
    # and words of the current sentence and the next sentence.
    apply_context(span, "words[-3:3] | sents[0:1]").text
    # Out: "sentence. A second sentence, longer this time. A third."

    # Will return the span covering at most the -5 and +5 words
    # around the span and the current sentence of the span.
    apply_context(span, "words[-4:4] & sent").text
    # Out: "A second sentence, longer this"
    ```

    !!! warning "Indexing"

        Unlike standard Python sequence slicing, `sents[0:0]` returns
        the current sentence, not an empty span.
    """

    @abc.abstractmethod
    def __call__(self, span: Span) -> Span:
        pass

    # logical ops
    def __and__(self, other: "ContextWindow"):
        # fmt: off
        return IntersectionContextWindow([
            *(self.contexts if isinstance(self, IntersectionContextWindow) else (self,)),  # noqa: E501
            *(other.contexts if isinstance(other, IntersectionContextWindow) else (other,))  # noqa: E501
        ])
        # fmt: on

    def __or__(self, other: "ContextWindow"):
        # fmt: off
        return UnionContextWindow([
            *(self.contexts if isinstance(self, UnionContextWindow) else (self,)),
            *(other.contexts if isinstance(other, UnionContextWindow) else (other,))
        ])
        # fmt: on

    @classmethod
    def parse(cls, query):
        try:
            return eval(
                query,
                {},
                {
                    "words": WordContextWindow,
                    "sents": SentenceContextWindow,
                    "sent": SentenceContextWindow(0, 0),
                },
            )
        except NameError:
            raise ValueError(
                "Only queries containing vars `words[before:after]`, "
                "`sents[before:after]` and `sent` are allowed to "
                f"define a context getter, got {query!r}"
            )

    @classmethod
    def validate(cls, obj, config=None):
        if isinstance(obj, cls):
            return obj
        if isinstance(obj, str):
            return cls.parse(obj)
        if isinstance(obj, tuple):
            assert len(obj) == 2
            return WordContextWindow(*obj)
        if isinstance(obj, int):
            assert obj != 0, "The provided `window` should not be 0"
            return WordContextWindow(obj, 0) if obj < 0 else WordContextWindow(0, obj)
        raise ValueError(f"Invalid context: {obj}")


class LeafContextWindowMeta(ContextWindowMeta):
    def __getitem__(cls, item) -> Span:
        assert isinstance(item, slice)
        before = item.start
        after = item.stop
        return cls(before, after)


class LeafContextWindow(ContextWindow, metaclass=LeafContextWindowMeta):
    pass


class WordContextWindow(LeafContextWindow):
    def __init__(
        self,
        before: Optional[int] = None,
        after: Optional[int] = None,
    ):
        self.before = before
        self.after = after

    def __call__(self, span):
        start = span.start + self.before if self.before is not None else 0
        end = span.end + self.after if self.after is not None else len(span.doc)
        return span.doc[max(0, start) : min(len(span.doc), end)]

    def __repr__(self):
        return "words[{}:{}]".format(self.before, self.after)


class SentenceContextWindow(LeafContextWindow):
    def __init__(
        self,
        before: Optional[int] = None,
        after: Optional[int] = None,
    ):
        self.before = before
        self.after = after

    def __call__(self, span):
        sent_starts = span.doc.to_array("SENT_START") == 1
        sent_indices = sent_starts.cumsum()
        sent_indices = sent_indices - sent_indices[span.start]

        start_idx = end_idx = None
        if self.before is not None:
            start = sent_starts & (sent_indices == self.before)
            x = np.flatnonzero(start)
            start_idx = x[-1] if len(x) else 0

        if self.after is not None:
            end = sent_starts & (sent_indices == self.after + 1)
            x = np.flatnonzero(end)
            end_idx = x[0] - 1 if len(x) else len(span.doc)

        return span.doc[start_idx:end_idx]

    def __repr__(self):
        return "sents[{}:{}]".format(self.before, self.after)


class UnionContextWindow(ContextWindow):
    def __init__(
        self,
        contexts: AsList[ContextWindow],
    ):
        self.contexts = contexts

    def __call__(self, span):
        results = [context(span) for context in self.contexts]
        min_word = min([span.start for span in results])
        max_word = max([span.end for span in results])
        return span.doc[min_word:max_word]

    def __repr__(self):
        return " | ".join(repr(context) for context in self.contexts)


class IntersectionContextWindow(ContextWindow):
    def __init__(
        self,
        contexts: AsList[ContextWindow],
    ):
        self.contexts = contexts

    def __call__(self, span):
        results = [context(span) for context in self.contexts]
        min_word = max([span.start for span in results])
        max_word = min([span.end for span in results])
        return span.doc[min_word:max_word]

    def __repr__(self):
        return " & ".join(repr(context) for context in self.contexts)



"""`eds.ckd` pipeline"""

from typing import Optional

from loguru import logger
from spacy.tokens import Doc, Span

from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import SpanSetterArg
from edsnlp.pipes.core.contextual_matcher.models import FullConfig
from edsnlp.pipes.ner.disorders.base import DisorderMatcher

from .patterns import default_patterns


class CKDMatcher(DisorderMatcher):
    """

    The `eds.CKD` pipeline component extracts mentions of CKD (Chronic Kidney Disease).
    It will notably match:

    - Mentions of various diseases (see below)
    - Kidney transplantation
    - Chronic dialysis
    - Renal failure **from stage 3 to 5**. The stage is extracted by trying 3 methods:
        - Extracting the mentioned stage directly ("*IRC stade IV*")
        - Extracting the severity directly ("*IRC terminale*")
        - Extracting the mentioned GFR (DFG in french) ("*IRC avec DFG estimé à 30
          mL/min/1,73m2)*")

    ??? info "Details of the used patterns"
        ```{ .python .no-check }
        # fmt: off
        --8<-- "edsnlp/pipes/ner/disorders/CKD/patterns.py"
        # fmt: on
        ```

    Extensions
    ----------
    On each span `span` that match, the following attributes are available:

    - `span._.detailed_status`: set to None
    - `span._.assigned`: dictionary with the following keys, if relevant:
        - `stage`: mentioned renal failure stage
        - `status`: mentioned renal failure severity (e.g. modérée, sévère, terminale,
          etc.)
        - `dfg`: mentioned DFG

    Examples
    --------
    ```python
    import edsnlp, edsnlp.pipes as eds

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(eds.sentences())
    nlp.add_pipe(
        eds.normalizer(
            accents=True,
            lowercase=True,
            quotes=True,
            spaces=True,
            pollution=dict(
                information=True,
                bars=True,
                biology=True,
                doctors=True,
                web=True,
                coding=True,
                footer=True,
            ),
        ),
    )
    nlp.add_pipe(eds.ckd())
    ```

    Below are a few examples:

    --8<-- "docs/assets/fragments/ckd-examples.md"

    Parameters
    ----------
    nlp : Optional[PipelineProtocol]
        The pipeline
    name : Optional[str]
        The name of the component
    patterns: FullConfig
        The patterns to use for matching
    label : str
        The label to use for the `Span` object and the extension
    span_setter : SpanSetterArg
        How to set matches on the doc

    Authors and citation
    --------------------
    The `eds.ckd` component was developed by AP-HP's Data Science team with a
    team of medical experts, following the insights of the algorithm proposed
    by [@petitjean_2024].
    """

    def __init__(
        self,
        nlp: Optional[PipelineProtocol],
        name: str = "ckd",
        *,
        patterns: FullConfig = default_patterns,
        label: str = "ckd",
        span_setter: SpanSetterArg = {"ents": True, "ckd": True},
    ):
        super().__init__(
            nlp=nlp,
            name=name,
            label=label,
            patterns=patterns,
            span_setter=span_setter,
        )

    def classify_from_dfg(self, dfg_span: Optional[Span]):
        if dfg_span is None:
            return False
        try:
            dfg_value = float(dfg_span.text.replace(",", ".").strip())
        except ValueError:
            logger.trace(f"DFG value couldn't be extracted from {dfg_span.text}")
            return False

        return dfg_value < 60  # We keep only moderate to severe CKD

    def process(self, doc: Doc):
        for span in super().process(doc):
            if span._.source == "dialysis" and "chronic" not in span._.assigned.keys():
                continue

            if span._.source == "general":
                if {"stage", "status"} & set(span._.assigned.keys()):
                    yield span
                elif self.classify_from_dfg(span._.assigned.get("dfg", None)):
                    yield span
            else:
                yield span



import edsnlp


def test_read_write(blank_nlp, text, df_notes_pyspark):
    # line below is just to mix params to avoid running too many tests
    shuffle = "dataset" if blank_nlp.lang == "eds" else False

    reader = edsnlp.data.from_spark(
        df_notes_pyspark,
        converter="omop",
        nlp=blank_nlp,
        shuffle=shuffle,
    ).set_processing(backend="simple")
    doc = list(reader)[0]
    assert doc.text == text

    blank_nlp.add_pipe("eds.matcher", config={"terms": {"douleur": ["douleurs"]}})
    blank_nlp.add_pipe("eds.negation")
    docs = blank_nlp.pipe(reader)

    writer = edsnlp.data.to_spark(
        docs,
        converter="omop",
        span_attributes=["negation"],
        span_getter=["ents"],
    )
    res = writer.toPandas().to_dict(orient="records")
    assert len(res) == 20
    assert sum(len(r["entities"]) for r in res) == 20



import os

import pytest

try:
    import optuna
except ImportError:
    optuna = None

if optuna is None:
    pytest.skip("optuna not installed", allow_module_level=True)

import shutil

from edsnlp.tune import (
    tune,
)


def assert_results(output_dir):
    results_file = os.path.join(output_dir, "results_summary.txt")
    assert os.path.exists(results_file)

    with open(results_file, "r") as f:
        content = f.read()
        assert "Study Summary" in content
        assert "Best trial" in content
        assert "Value" in content
        assert "Params" in content
        assert "Importances" in content

    config_file = os.path.join(output_dir, "config.yml")
    assert os.path.exists(config_file), f"Expected file {config_file} not found"

    optimization_history_file = os.path.join(output_dir, "optimization_history.html")
    assert os.path.exists(optimization_history_file), (
        f"Expected file {optimization_history_file} not found"
    )

    parallel_coord_file = os.path.join(output_dir, "parallel_coordinate.html")
    assert os.path.exists(parallel_coord_file), (
        f"Expected file {parallel_coord_file} not found"
    )

    contour_file = os.path.join(output_dir, "contour.html")
    assert os.path.exists(contour_file), f"Expected file {contour_file} not found"

    edf_file = os.path.join(output_dir, "edf.html")
    assert os.path.exists(edf_file), f"Expected file {edf_file} not found"

    timeline_file = os.path.join(output_dir, "timeline.html")
    assert os.path.exists(timeline_file), f"Expected file {timeline_file} not found"


@pytest.mark.parametrize(
    "two_phase_tuning,n_trials,start_from_checkpoint",
    [
        (False, None, False),
        (False, 7, False),
        (False, 7, True),
        (True, None, False),
        (True, 7, False),
        (True, 7, True),
    ],
)
def test_tune(tmpdir, two_phase_tuning, n_trials, start_from_checkpoint):
    config_meta = {"config_path": ["tests/tuning/config.yml"]}
    hyperparameters = {
        "optimizer.groups.'.*'.lr.start_value": {
            "alias": "start_value",
            "type": "float",
            "low": 1e-4,
            "high": 1e-3,
            "log": True,
        },
        "optimizer.groups.'.*'.lr.warmup_rate": {
            "alias": "warmup_rate",
            "type": "float",
            "low": 0.0,
            "high": 0.3,
            "step": 0.05,
        },
    }
    output_dir = "./results"
    checkpoint_dir = "./tests/tuning/test_checkpoints"
    try:
        gpu_hours = 0.015 if n_trials is None else None
        seed = 42
        metric = "ner.micro.f"
        os.makedirs(checkpoint_dir, exist_ok=True)
        if start_from_checkpoint:
            assert gpu_hours is None
            test_schedule = [(n_trials, True), (n_trials, False)]
        else:
            test_schedule = [(n_trials, False)]
        for n_trials_, keep_checkpoint in test_schedule:
            tune(
                config_meta=config_meta,
                hyperparameters=hyperparameters,
                output_dir=output_dir,
                checkpoint_dir=checkpoint_dir,
                gpu_hours=gpu_hours,
                n_trials=n_trials_,
                two_phase_tuning=two_phase_tuning,
                seed=seed,
                metric=metric,
                keep_checkpoint=keep_checkpoint,
            )
        if two_phase_tuning:
            phase_1_dir = os.path.join(output_dir, "phase_1")
            phase_2_dir = os.path.join(output_dir, "phase_2")
            if not start_from_checkpoint:
                assert_results(phase_1_dir)
            assert_results(phase_2_dir)
        else:
            assert_results(output_dir)
    finally:
        shutil.rmtree(output_dir, ignore_errors=True)
        shutil.rmtree(checkpoint_dir, ignore_errors=True)
        shutil.rmtree("./artifacts", ignore_errors=True)



from itertools import islice
from pathlib import Path

import polars
import pytest

import edsnlp


def test_read_write(blank_nlp, text, df_notes_pandas):
    import polars

    df_notes_polars = polars.from_pandas(df_notes_pandas)
    reader = edsnlp.data.from_polars(
        df_notes_polars,
        converter="omop",
        nlp=blank_nlp,
    ).set_processing(backend="simple")
    doc = list(reader)[0]
    assert doc.text == text

    blank_nlp.add_pipe("eds.matcher", config={"terms": {"douleur": ["douleurs"]}})
    blank_nlp.add_pipe("eds.negation")
    docs = reader.map_pipeline(blank_nlp)

    writer: polars.DataFrame = docs.to_polars(
        converter="omop",
        span_attributes=["negation"],
        span_getter=["ents"],
    )
    res = writer.to_dicts()
    assert len(res) == 20
    assert sum(len(r["entities"]) for r in res) == 20


@pytest.mark.parametrize("num_cpu_workers", [0, 2])
def test_read_shuffle_loop(num_cpu_workers: int):
    data = polars.read_parquet(
        Path(__file__).parent.parent.resolve() / "resources" / "docs.parquet"
    )
    notes_a, notes_b = (
        edsnlp.data.from_polars(
            data,
            shuffle="dataset",
            seed=42,
            loop=True,
        )
        .map(lambda x: x["note_id"])
        .set_processing(num_cpu_workers=num_cpu_workers)
        for _ in range(2)
    )
    # This test differs from other data rand perm test as polars rng has changed
    # between versions (from 1.32 ?) so it's easier to check this
    notes_a = list(islice(notes_a, 6))
    notes_b = list(islice(notes_b, 6))
    assert notes_a == notes_b, "Shuffling with loop should yield the same results"



import copyreg
import math
import warnings
from enum import Enum
from typing import TypeVar

import dill
import torch

# filter "is in beta" torch warnings
warnings.filterwarnings("ignore", message=".*is in beta.*")

Args = TypeVar("Args")


def pad_2d(data, pad=0, **kwargs):
    max_len = max(map(len, data), default=0)
    padded = [row + [pad] * (max_len - len(row)) for row in data]
    return torch.as_tensor(padded, **kwargs)


class ActivationFunction(str, Enum):
    relu = "relu"
    gelu = "gelu"
    glu = "glu"


def get_activation_function(activation: ActivationFunction):
    return getattr(torch.nn.functional, activation)


def mask_to_triangle(mask):
    """
    Convert a boolean mask to a tensor containing distance to the nearest
    edge of the mask, in each direction.
    For example, if the mask is
    ```
    [
        [1, 1, 1, 1, 1],
        [1, 1, 1, 0, 0],
    ]
    ```

    The output will be
    ```
    [
        [0, 1, 2, 1, 0],
        [0, 1, 0, -1, -2]
    ]
    ```

    Parameters
    ----------
    mask: torch.Tensor
        A boolean mask

    Returns
    -------
    torch.Tensor
    """
    ramp = torch.arange(0, mask.shape[1], 1)
    scores = torch.min(ramp, mask.sum(1, keepdim=True) - 1 - ramp).view(-1)
    return scores


def make_windows(lengths, size, stride):
    max_len = max(lengths)
    windows = pad_2d(
        [
            list(
                range(
                    idx * stride + doc_idx * max_len,
                    min(idx * stride + size, length) + doc_idx * max_len,
                )
            )
            for doc_idx, length in enumerate(lengths)
            for idx in range(0, 1 + max(0, math.ceil((length - size) / stride)))
        ],
        pad=-1,
        dtype=torch.long,
    )
    windows_mask = windows != -1
    windows[~windows_mask] = 0
    indexer = torch.zeros((len(lengths), max_len), dtype=torch.long).view(-1)
    scores = mask_to_triangle(windows_mask)
    scores = scores * len(scores) + torch.arange(len(scores))
    scores[~windows_mask.view(-1)] = -1
    indexer.index_reduce_(
        dim=0,
        source=scores,
        index=windows.view(-1),
        reduce="amax",
    )
    indexer %= len(scores)
    return windows.masked_fill(~windows_mask, -1), indexer


def reduce_empty(*args, **kwargs):
    return type(None), ()


def load_pruned_obj(obj, _):
    return obj


# Torch may still be imported as a namespace package, so we can access the
# torch.save and torch.load functions

MAP_LOCATION = None


try:
    from accelerate.hooks import AlignDevicesHook

    # We need to replace the "execution_device" attribute of the AlignDevicesHook
    # using map_location when unpickling the stream

    def save_align_devices_hook(pickler, obj):
        pickler.save_reduce(load_align_devices_hook, (obj.__dict__,), obj=obj)

    def load_align_devices_hook(state):
        state["execution_device"] = MAP_LOCATION
        new_obj = AlignDevicesHook.__new__(AlignDevicesHook)
        new_obj.__dict__.update(state)
        return new_obj

except ImportError:
    AlignDevicesHook = None


def dump(
    *args,
    skip_tensors: bool = False,
    **kwargs,
):
    # We need to replace the "execution_device" attribute of the AlignDevicesHook
    # using map_location when pickling the stream
    old = None
    old_settings = dict(dill.settings)
    old_dispatch = {}
    try:
        if skip_tensors:
            if torch.Tensor in copyreg.dispatch_table:
                old_dispatch[torch.Tensor] = copyreg.dispatch_table[torch.Tensor]
            copyreg.pickle(torch.Tensor, reduce_empty)
        if AlignDevicesHook is not None:
            old = dill.Pickler.dispatch.get(AlignDevicesHook)
            dill.Pickler.dispatch[AlignDevicesHook] = save_align_devices_hook
        dill.settings["recurse"] = False
        dill.settings["byref"] = True
        return torch.save(*args, pickle_module=dill, **kwargs)
    finally:
        dill.settings.update(old_settings)
        if AlignDevicesHook is not None:
            del dill.Pickler.dispatch[AlignDevicesHook]
            if old is not None:  # pragma: no cover
                dill.Pickler.dispatch[AlignDevicesHook] = old
        copyreg.dispatch_table.pop(torch.Tensor, None)
        copyreg.dispatch_table.update(old_dispatch)


def load(*args, map_location=None, **kwargs):
    global MAP_LOCATION
    MAP_LOCATION = map_location
    if torch.__version__ >= "2.1" and isinstance(args[0], str):
        kwargs["mmap"] = True
    try:
        if torch.__version__ < "2.0.0":
            torch.load.__globals__["pickle"] = dill
        result = torch.load(
            *args,
            pickle_module=dill,
            map_location=map_location,
            **kwargs,
        )
    finally:
        import pickle

        torch.load.__globals__["pickle"] = pickle
    MAP_LOCATION = None
    return result



results_solid_tumor = dict(
    has_match=[True, True, False, True, True, True, True, True, True],
    detailled_status=[
        "LOCALIZED",
        "LOCALIZED",
        None,
        "METASTASIS",
        "METASTASIS",
        "LOCALIZED",
        "METASTASIS",
        "METASTASIS",
        "METASTASIS",
        "METASTASIS",
    ],
    assign=None,
    texts=[
        "Présence d'un carcinome intra-hépatique.",
        "Patient avec un K sein.",
        "Il y a une tumeur bénigne",
        "Tumeur métastasée",
        "Cancer du poumon au stade 4",
        "Cancer du poumon au stade 2",
        "Présence de nombreuses lésions secondaires",
        "Patient avec fracture abcddd secondaire. Cancer de",
        "Patient avec lesions non ciblées",
        "TNM: pTx N1 M1",
    ],
)

solid_tumor_config = dict(use_patterns_metastasis_ct_scan=True, use_tnm=True)



from typing import List

from pytest import fixture, mark
from spacy.tokens import Span

from edsnlp.pipes.qualifiers.family import FamilyContext
from edsnlp.utils.examples import parse_example

examples: List[str] = [
    (
        "Le père est <ent family=True>asthmatique</ent>, "
        "sans traitement traitement particulier."
    ),
    "Son père est atteint de la <ent family=True>COVID</ent>",
    "Son père a une infection au <ent family=True>COVID</ent>",
    "Son père a une possible infection au <ent family=True>COVID</ent>",
    (
        "Le père du patient a eu un <ent family_=FAMILY>cancer du colon</ent>. "
        "La mère se porte bien."
    ),
    "Antécédents familiaux : <ent family_=FAMILY>diabète</ent>.",
    "Un <ent family_=PATIENT>relevé</ent> sanguin a été effectué.",
    (
        "Antécédent familiaux de diabète mais pas "
        "<ent family_=PATIENT>détecté</ent> jusqu'ici."
    ),
    "mère : <ent family=True>diabète de type II</ent>",
]


@fixture
def family_factory(blank_nlp):
    default_config = dict(
        family=None,
        termination=None,
        attr="NORM",
        use_sections=False,
        explain=True,
    )

    def factory(on_ents_only, **kwargs):
        config = dict(**default_config)
        config.update(kwargs)

        return FamilyContext(
            nlp=blank_nlp,
            on_ents_only=on_ents_only,
            **config,
        )

    return factory


@mark.parametrize("on_ents_only", [True, False])
@mark.parametrize("use_sections", [True, False])
def test_family(blank_nlp, family_factory, on_ents_only, use_sections):
    family = family_factory(
        on_ents_only=on_ents_only,
        use_sections=use_sections,
    )

    for example in examples:
        text, entities = parse_example(example=example)

        doc = blank_nlp(text)
        doc.ents = [
            doc.char_span(ent.start_char, ent.end_char, label="ent") for ent in entities
        ]

        doc = family(doc)

        for entity, ent in zip(entities, doc.ents):
            for modifier in entity.modifiers:
                assert bool(ent._.family_cues) == (modifier.value in {"FAMILY", True})

                assert getattr(ent._, modifier.key) == modifier.value, (
                    f"{modifier.key} labels don't match."
                )

                if not on_ents_only:
                    for token in ent:
                        assert getattr(token._, modifier.key) == modifier.value, (
                            f"{modifier.key} labels don't match."
                        )


def test_on_span(blank_nlp, family_factory):
    doc = blank_nlp("Le père du patient est asthmatique, le patient n'est pas malade.")
    doc.ents = [
        Span(doc, 5, 6, label="ent"),
        Span(doc, 12, 13, label="ent"),
    ]

    family = family_factory(on_ents_only=True)
    res = family.process(doc[1:13])
    assert [(ent.ent.text, bool(ent.family)) for ent in res.ents] == [
        ("asthmatique", True),
        ("malade", False),
    ]

    family = family_factory(on_ents_only=False)
    res = family.process(doc[1:13])
    assert [(t.token.text, t.family) for t in res.tokens] == [
        ("père", True),
        ("du", True),
        ("patient", True),
        ("est", True),
        ("asthmatique", True),
        (",", False),
        ("le", False),
        ("patient", False),
        ("n'", False),
        ("est", False),
        ("pas", False),
        ("malade", False),
    ]



import warnings
from dataclasses import dataclass
from itertools import chain
from typing import Dict, List, Optional, Set, Union

from spacy.tokens import Doc, Span, Token

from edsnlp.core import PipelineProtocol
from edsnlp.matchers.phrase import EDSPhraseMatcher
from edsnlp.matchers.regex import RegexMatcher
from edsnlp.pipes.base import (
    BaseSpanAttributeClassifierComponent,
    SpanGetterArg,
    validate_span_getter,
)


def check_normalizer(nlp: PipelineProtocol) -> None:
    components = {name: component for name, component in nlp.pipeline}
    normalizer = components.get("normalizer")

    if normalizer and not normalizer.lowercase:
        warnings.warn(
            "You have chosen the NORM attribute, but disabled lowercasing "
            "in your normalisation pipeline. "
            "This WILL hurt performance : you might want to use the "
            "LOWER attribute instead."
        )


@dataclass
class BaseTokenQualifierResults:
    """
    Base dataclass to store qualification informations for each token.
    Specific qualifier pipes should inherit from this class and
    add specific fields as needed
    """

    token: Token


@dataclass
class BaseEntQualifierResults:
    """
    Base dataclass to store qualification informations for each entity.
    Specific qualifier pipes should inherit from this class and
    add specific fields as needed
    """

    ent: Span


@dataclass
class BaseQualifierResults:
    """
    Base dataclass to store qualification informations for all tokens and entities.
    """

    tokens: List[BaseTokenQualifierResults]
    ents: List[BaseEntQualifierResults]


class RuleBasedQualifier(BaseSpanAttributeClassifierComponent):
    """
    Implements the ConText algorithm (eq. NegEx for negations) for detecting contextual
    attributes text.

    Parameters
    ----------
    nlp : PipelineProtocol
        The pipeline object.
    attr : str
        spaCy's attribute to use:
        a string with the value "TEXT" or "NORM", or a dict with the key 'term_attr'
        we can also add a key for each regex.
    span_getter : SpanGetterArg
        Which entities should be classified. By default, `doc.ents`
    on_ents_only : Union[bool, str, List[str], Set[str]]
        Whether to look for matches around detected entities only.
        Useful for faster inference in downstream tasks.

        - If True, will look in all ents located in `doc.ents` only
        - If an iterable of string is passed, will additionally look in `doc.spans[key]`
        for each key in the iterable
    explain : bool
        Whether to keep track of cues for each entity.
    **terms : Dict[str, Optional[List[str]]]
        Terms to look for.
    """

    defaults = dict()

    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = None,
        *,
        attr: str,
        attributes: List[str],
        span_getter: SpanGetterArg,
        on_ents_only: Union[bool, str, List[str], Set[str]],
        explain: bool,
        terms: Dict[str, Optional[List[str]]],
        regex: Dict[str, Optional[List[str]]] = {},
    ):
        if attr.upper() == "NORM":
            check_normalizer(nlp)

        self.phrase_matcher = EDSPhraseMatcher(vocab=nlp.vocab, attr=attr)
        self.phrase_matcher.build_patterns(nlp=nlp, terms=terms)

        self.regex_matcher = RegexMatcher(attr=attr)
        self.regex_matcher.build_patterns(regex=regex)
        self.attributes = attributes

        self.on_ents_only = on_ents_only

        if span_getter is None and on_ents_only is None:
            on_ents_only = True

        if on_ents_only:
            assert span_getter is None or on_ents_only is True, (
                "Cannot use both `span_getter` and `on_ents_only` as a span selection "
                "argument."
            )
            assert isinstance(on_ents_only, (list, str, set, bool)), (
                "The `on_ents_only` argument should be a "
                "string, a bool, a list or a set of string"
            )
            span_getter = "ents" if on_ents_only is True else on_ents_only
        else:
            span_getter = "ents"
        self.explain = explain
        super().__init__(
            nlp=nlp,
            name=name,
            span_getter=validate_span_getter(span_getter),
        )

    def get_cues(self, doc: Doc, spans: Optional[List[Span]] = None):
        """
        Extract cues (ex: ne/pas for negations) from the document.

        Parameters
        ----------
        doc: Doc
            The document to process.
        spans: Optional[List[Span]]
            Optional list of spans to limit the search around.
            If None, will search in the whole document.

        Returns
        -------
        List[Span]
            List of detected cues
        """
        if spans is None:
            matches = (
                *self.phrase_matcher(doc, as_spans=True),
                *self.regex_matcher(doc, as_spans=True),
            )
        else:
            sents = dict.fromkeys((span.sent for span in spans))
            matches = chain.from_iterable(
                (
                    (
                        *self.phrase_matcher(s, as_spans=True),
                        *self.regex_matcher(s, as_spans=True),
                    )
                    for s in sents
                )
            )
        return list(matches)

    def process(self, doc_like: Union[Doc, Span]) -> BaseQualifierResults:
        raise NotImplementedError

    def __call__(self, doc: Doc) -> Doc:
        results = self.process(doc)  # pragma: no cover
        raise NotImplementedError(f"{type(results)} should be used to tag the document")



import re
from contextlib import contextmanager
from types import SimpleNamespace


@contextmanager
def mock_llm_service(responder, *, base_url="http://localhost:8080/v1"):
    """
    Patch the OpenAI client so calls to chat.completions.create go through
    responder instead.
    """

    import inspect
    from unittest.mock import patch

    from openai.resources.chat.completions.completions import (
        AsyncCompletions,
        Completions,
    )

    tag_pattern = re.compile(r"</?[^>]+?>")

    def extract_user_text(messages):
        if not messages:
            return None
        for message in reversed(messages):
            if message.get("role") != "user":
                continue
            content = message.get("content", "")
            if isinstance(content, list):
                return "".join(
                    part.get("text", "")
                    for part in content
                    if isinstance(part, dict) and part.get("type") == "text"
                )
            return str(content)
        return None

    def to_openai_response(result, *, messages):
        if inspect.isawaitable(result):  # pragma: no cover
            raise TypeError("responder must return a string or response, not awaitable")

        if isinstance(result, str):
            content = result
            user_text = extract_user_text(messages)
            if user_text:
                plain_fragment = tag_pattern.sub("", content)
                if plain_fragment and plain_fragment in user_text:
                    content = user_text.replace(plain_fragment, content, 1)
            return SimpleNamespace(
                choices=[
                    SimpleNamespace(
                        message=SimpleNamespace(content=content),
                    )
                ]
            )

        if isinstance(result, dict) and "choices" in result:
            return SimpleNamespace(
                choices=[
                    SimpleNamespace(
                        message=SimpleNamespace(content=choice["message"]["content"])
                    )
                    for choice in result["choices"]
                ]
            )

        return result

    def check_base_url(client) -> None:
        if base_url is None:
            return
        try:
            client_url = str(client.base_url)
        except AttributeError:
            try:
                client_url = str(client._client.base_url)
            except AttributeError:  # pragma: no cover - very defensive
                return
        assert client_url.rstrip("/") == base_url.rstrip("/"), (
            f"Unexpected base_url {client_url}, expected {base_url}"
        )

    def sync_create(self, *args, **kwargs):
        if args:
            raise TypeError("mocked completions.create expects keyword arguments only")
        check_base_url(self._client)
        return to_openai_response(responder(**kwargs), messages=kwargs.get("messages"))

    async def async_create(self, *args, **kwargs):
        if args:
            raise TypeError("mocked completions.create expects keyword arguments only")
        check_base_url(self._client)
        result = responder(**kwargs)
        if inspect.isawaitable(result):
            result = await result
        return to_openai_response(result, messages=kwargs.get("messages"))

    with patch.object(Completions, "create", sync_create), patch.object(
        AsyncCompletions, "create", async_create
    ):
        yield



from confit import Cli

from edsnlp.training.trainer import *  # noqa: F403
from edsnlp.training.trainer import registry, train

app = Cli(pretty_exceptions_enable=False)
train_command = app.command(name="train", registry=registry)(train)

if __name__ == "__main__":
    app()



import ast
import inspect
import re
import sys
import textwrap
import warnings
from math import isclose

import catalogue
import pytest
from spacy.tokens.underscore import Underscore

pytest.importorskip("mkdocs")
try:
    import torch.nn
except ImportError:
    torch = None

if torch is None:
    pytest.skip("torch not installed", allow_module_level=True)
pytest.importorskip("rich")

from extract_docs_code import extract_docs_code  # noqa: E402

# We don't check documentation for Python <= 3.7:
if sys.version_info < (3, 8):
    url_to_code = {}
else:
    url_to_code = dict(extract_docs_code())
    # just to make sure something didn't go wrong
    assert len(url_to_code) > 50


class nested_approx:
    def __init__(self, value, rel=1e-12, abs=1e-12):
        self._value, self._rel, self._abs = value, rel, abs

    def __eq__(self, other):
        return self._match(self._value, other)

    def __req__(self, other):
        return self._match(other, self._value)

    __hash__ = None  # keep it un-hashable

    def _match(self, a, b):
        if isinstance(a, (int, float)) and isinstance(b, (int, float)):
            return isclose(a, b, rel_tol=self._rel, abs_tol=self._abs)
        if isinstance(a, (list, tuple)):
            return (
                isinstance(b, (list, tuple))
                and len(a) == len(b)
                and all(self._match(x, y) for x, y in zip(a, b))
            )
        if isinstance(a, dict):
            return (
                isinstance(b, dict)
                and a.keys() == b.keys()
                and all(self._match(a[k], b[k]) for k in a)
            )
        return a == b

    def __repr__(self):
        return f"nested_approx({self._value!r}, rel={self._rel}, abs={self._abs})"


pytest.nested_approx = nested_approx


def printer(code: str) -> None:
    """
    Prints a code bloc with lines for easier debugging.

    Parameters
    ----------
    code : str
        Code bloc.
    """
    lines = []
    for i, line in enumerate(code.split("\n")):
        lines.append(f"{i + 1:03}  {line}")

    print("\n".join(lines))


def insert_assert_statements(code):
    line_table = [0]
    for line in code.splitlines(keepends=True):
        line_table.append(line_table[-1] + len(line))

    tree = ast.parse(code)
    replacements = []

    for match in re.finditer(
        r"^\s*#\s*Out\s*: (.*$(?:\n#\s.*$)*)", code, flags=re.MULTILINE
    ):
        lineno = code[: match.start()].count("\n")
        for stmt in tree.body:
            if stmt.end_lineno == lineno:
                if isinstance(stmt, ast.Expr):
                    expected = textwrap.dedent(match.group(1)).replace("\n# ", "\n")
                    expected_s = expected
                    begin = line_table[stmt.lineno - 1]
                    if not (expected.startswith("'") or expected.startswith('"')):
                        expected_s = repr(expected)
                    end = match.end()
                    stmt_str = ast.unparse(stmt)
                    if stmt_str.startswith("print("):
                        stmt_str = stmt_str.replace("print", "print_to_string")
                    repl = f"""\
val = {stmt_str}
try:
    import ast
    expected = ast.literal_eval({expected_s})
except (ValueError, SyntaxError):
    expected = None
assert str(val) == {expected_s} or val == pytest.nested_approx(expected, 0.01, 0.01)
"""
                    replacements.append((begin, end, repl))
                if isinstance(stmt, ast.For):
                    expected = textwrap.dedent(match.group(1)).split("\n# Out: ")
                    expected = [line.replace("\n# ", "\n") for line in expected]
                    begin = line_table[stmt.lineno - 1]
                    end = match.end()
                    stmt_str = ast.unparse(stmt).replace("print", "assert_print")
                    repl = f"""\
printed = []
{stmt_str}
assert printed == {expected}
"""
                    replacements.append((begin, end, repl))

    for begin, end, repl in reversed(replacements):
        code = code[:begin] + repl + code[end:]

    return code


# TODO: once in a while, it can be interesting to run reset_imports for each code block,
#  instead of only once and tests should still pass, but it's way slower.
@pytest.fixture(scope="module")
def reset_imports():
    """
    Reset the imports for each test.
    """
    # 1. Clear registered functions to avoid using cached ones
    for k, m in list(catalogue.REGISTRY.items()):
        mod = inspect.getmodule(m)
        if mod is not None and mod.__name__.startswith("edsnlp"):
            del catalogue.REGISTRY[k]

    # Let's ensure that we "bump" into every possible warnings:
    # 2. Remove all modules that start with edsnlp, to reimport them
    for k in list(sys.modules):
        if k.split(".")[0] == "edsnlp":
            del sys.modules[k]

    # 3. Delete spacy extensions to avoid error when re-importing
    Underscore.span_extensions.clear()
    Underscore.doc_extensions.clear()
    Underscore.token_extensions.clear()


# Note the use of `str`, makes for pretty output
@pytest.mark.parametrize("url", sorted(url_to_code.keys()), ids=str)
def test_code_blocks(url, tmpdir, reset_imports):
    code = url_to_code[url]
    code_with_asserts = """
import pytest

def print_to_string(*args, sep=" ", end="\\n", file=None, flush=False):
    return (sep.join(map(str, args)) + end).rstrip('\\n')

def assert_print(*args, sep=" ", end="\\n", file=None, flush=False):
    printed.append(print_to_string(*args, sep=sep, end=end, file=file, flush=flush))

""" + insert_assert_statements(code)
    assert "# Out:" not in code_with_asserts, (
        "Unparsed asserts in {url}:\n" + code_with_asserts
    )
    # We'll import test_code_blocks from here
    sys.path.insert(0, str(tmpdir))
    test_file = tmpdir.join("test_code_blocks.py")

    # Clear all warnings
    warnings.resetwarnings()

    try:
        with warnings.catch_warnings():
            warnings.simplefilter("error")
            warnings.filterwarnings(module=".*endlines.*", action="ignore")
            warnings.filterwarnings(
                message="__package__ != __spec__.parent", action="ignore"
            )
            # First, forget test_code_blocks
            sys.modules.pop("test_code_blocks", None)

            # Then, reimport it, to let pytest do its assertion rewriting magic
            test_file.write_text(code_with_asserts, encoding="utf-8")

            import test_code_blocks  # noqa: F401

            exec(
                compile(code_with_asserts, test_file, "exec"),
                {"__MODULE__": "__main__"},
            )
    except ImportError as e:
        # skip UMLS tests if the package is not installed
        if "umls_downloader" in str(e):
            pytest.skip("umls_downloader package not found")
        else:
            raise
    except Exception:
        printer(code_with_asserts)
        raise



from datetime import datetime
from functools import partial

import spacy
from pytest import mark
from spacy.tokens.span import Span

from edsnlp.pipes.qualifiers.history import History
from edsnlp.pipes.qualifiers.history.patterns import history
from edsnlp.pipes.terminations import termination

text = """COMPTE RENDU D'HOSPITALISATION du 11/07/2018 au 12/07/2018

MOTIF D'HOSPITALISATION
Monsieur Dupont 30\n2 Jean Michel, de sexe masculin, âgée de 39 ans,
née le 23/11/1978, est admis pour une toux.
Il a été hospitalisé du 11/08/2019 au 17/08/2019,
avec un antécédent d'asthme il y a 25 jours.

ANTÉCÉDENTS
Antécédents médicaux :
Premier épisode: il a été hospitalisé pour asthme cette semaine-ci,
il y a 3 jours, le 13 août 2020.
Hier, le patient est venu pour une toux dont les symptômes,
seraient apparus il y a 2 mois.
L'asthme est critique

MODE DE VIE
Rien à signaler.
"""


@mark.parametrize("use_sections", [True, False])
@mark.parametrize("use_dates", [True, False])
@mark.parametrize("on_ents_only", [True, False])
@mark.parametrize("exclude_birthdate", [True, False])
def test_history(lang, use_sections, use_dates, exclude_birthdate, on_ents_only):
    nlp = spacy.blank(lang)
    nlp.add_pipe("eds.sentences")
    nlp.add_pipe("eds.normalizer")
    nlp.add_pipe(
        "eds.matcher",
        config=dict(
            terms=dict(
                respiratoire=[
                    "asthmatique",
                    "asthme",
                    "toux",
                ]
            )
        ),
    )
    nlp.add_pipe(
        "eds.history", config=dict(use_sections=use_sections, use_dates=use_dates)
    )
    doc = nlp(text)
    nlp.remove_pipe("eds.history")
    nlp.add_pipe("eds.sections")
    nlp.add_pipe("eds.dates")
    doc = nlp(text)
    doc._.note_datetime = datetime(2020, 8, 11)
    doc._.birth_datetime = datetime(1978, 11, 23)

    history_nlp = History(
        nlp=nlp,
        attr="NORM",
        history=history,
        termination=termination,
        use_sections=use_sections,
        use_dates=use_dates,
        exclude_birthdate=exclude_birthdate,
        closest_dates_only=True,
        history_limit=15,
        explain=True,
        on_ents_only=on_ents_only,
    )
    doc = history_nlp(doc)
    assert [t.text for t in doc.ents] == ["toux", "asthme", "asthme", "toux", "asthme"]

    if use_dates:
        assert doc.ents[0]._.history is not exclude_birthdate
        if not exclude_birthdate:
            assert doc.ents[0]._.history_cues[0].text == "23/11/1978"

    assert doc.ents[1]._.history and doc.ents[1]._.history_cues[0].label_ == "history"

    if use_sections:
        assert doc.ents[2]._.history is not use_dates
        assert doc.ents[2]._.history_cues[0].label_ == "antécédents"
        if use_dates:
            assert doc.ents[2]._.recent_cues[0].label_ == "relative_date"
        # If we use date, the first asthme entity of the ANTÉCÉDENTS section
        # will be considered to recent => not history. Otherwise, it's in the
        # ANTÉCÉDENTS section, so it is considered history.
        assert doc.ents[2]._.history == (not use_dates)
        # Same as above, but now the date is "Hier"
        assert doc.ents[3]._.history == (not use_dates)
        assert doc.ents[4]._.history  # Default behavior of ANTÉCÉDENTS section


def test_on_span(blank_nlp):
    doc = blank_nlp(
        "Lésion pulmonaire avec antécédent de lésion secondaire associée, et voilà."
    )
    doc.ents = [Span(doc, 5, 7, label="lesion")]
    history_factory = partial(
        History,
        nlp=blank_nlp,
        attr="NORM",
        use_sections=True,
        use_dates=True,
        exclude_birthdate=False,
        closest_dates_only=True,
        history_limit=15,
        explain=True,
        on_ents_only=True,
    )

    history = history_factory(on_ents_only=True)
    res = history.process(doc[2:11])
    assert [(str(ent.ent), ent.history) for ent in res.ents] == [
        ("lésion secondaire", True)
    ]

    history = history_factory(on_ents_only=False)
    res = history.process(doc[2:8])
    assert [(t.token.text, t.history) for t in res.tokens] == [
        ("avec", True),
        ("antécédent", True),
        ("de", True),
        ("lésion", True),
        ("secondaire", True),
        ("associée", True),
        (",", False),
        ("et", False),
        ("voilà", False),
    ]



# ruff: noqa: E501
"""
This module is meant to be run as a script to test multiprocessing
with a pipe declared in the main module.
Without the dill settings in multiprocessing.py
```
    dill.settings["recurse"] = False
    dill.settings["byref"] = True
```
this fails with an error like
```
_pickle.PicklingError: Can't pickle <cyfunction ValidatedFunction.create_model.<locals>.DecoratorBaseModel.check_args at 0x...>:
it's not found as pydantic.decorator.ValidatedFunction.create_model.<locals>.DecoratorBaseModel.check_args
```
"""

from typing import Optional

from spacy.tokens import Doc

import edsnlp
from edsnlp import registry
from edsnlp.core import PipelineProtocol
from edsnlp.pipes.base import BaseComponent


class SimplePipe(BaseComponent):
    def __init__(
        self,
        nlp: PipelineProtocol,
        name: Optional[str] = "simple_pipe",
    ):
        super().__init__(nlp=nlp, name=name)

    def __call__(self, doc: Doc) -> Doc:
        return doc


if __name__ == "__main__":
    create_component = registry.factory.register(
        "test_simple_pipe",
        assigns=["doc.spans"],
    )(SimplePipe)

    nlp = edsnlp.blank("eds")
    nlp.add_pipe(create_component())

    docs = [f"Texte {i}" for i in range(100)]
    docs = nlp.pipe(docs)
    docs = docs.set_processing(backend="multiprocessing")
    list(docs)



import json
from pathlib import Path
from typing import Dict

import torch
from confit import Cli

import edsnlp
from edsnlp.core.registries import registry
from edsnlp.train import GenericScorer, SampleGenerator

app = Cli(pretty_exceptions_show_locals=False)


@app.command(name="evaluate", registry=registry)
def evaluate(
    *,
    data: SampleGenerator,
    model_path: Path = "artifacts/model-last",
    scorer: GenericScorer,
    task_metadata: Dict = {},
):
    """
    Evaluate a model on a dataset. This function can be called from the command line or
    from a script.

    By default, the model is loaded from `artifacts/model-last`, and the results are
    stored both in `artifacts/test_metrics.json` and in the model's
    `artifacts/model-last/meta.json` file.

    Parameters
    ----------
    data: SampleGenerator
        A function that generates samples for evaluation
    model_path: Path
        The path to the model to evaluate
    scorer: GenericScorer
        A function that computes metrics on the model. You can also pass a dict:

        ```{ .python .no-check }
        scorer = {
            "ner": NerExactMetric(...),
            ...
        }
        ```
    task_metadata: Dict
        Metadata about the evaluation task. This will be stored in the model's meta.json
        file and is primarily meant to be parsed by the Hugging Face Hub, e.g.,
        but also to remove previous results for the same dataset.

        ```{ .python .no-check }
        task_metadata={
            "task": {"type": "token-classification"},
            "dataset": {
                "name": "my_dataset",
                "type": "private",
            },
        }
        ```

    Returns
    -------

    """
    nlp = edsnlp.load(model_path).to("cuda" if torch.cuda.is_available() else "cpu")
    val_docs = list(data(nlp))
    metrics = scorer(nlp, val_docs)

    print("Metrics:")
    print(json.dumps(metrics, indent=2))
    Path("artifacts/test_metrics.json").write_text(json.dumps(metrics, indent=2))
    meta = json.loads((model_path / "meta.json").read_text())
    results = meta.setdefault("results", [])

    # Remove previous results for the same dataset
    dataset_name = task_metadata.get("dataset", {}).get("name")
    index = next(
        (
            i
            for i, res in enumerate(results)
            if res.get("dataset", {}).get("name") == dataset_name
        ),
        None,
    )
    if index is not None:
        results.pop(index)
    results.append(
        {
            "metrics": metrics,
            **task_metadata,
        }
    )
    (model_path / "meta.json").write_text(json.dumps(meta, indent=2))
    Path("artifacts/test_metrics.jsonl").write_text(json.dumps(metrics, indent=2))


if __name__ == "__main__":
    app()



from spacy.tokens import Doc
from spacy.tokens.span import Span

import edsnlp
import edsnlp.pipes as eds


def test_split_line_jump():
    txt = """This is a test. Another test.

A third test!"""
    nlp = edsnlp.blank("eds")
    nlp.add_pipe("eds.matcher", config={"terms": {"test": "test"}})
    doc = nlp(txt)
    Span.set_extension("test_dict", default={}, force=True)
    Doc.set_extension("global_attr", default=None, force=True)
    Span.set_extension("ent_attr", default=None, force=True)
    doc._.global_attr = "global"
    doc.ents[0]._.test_dict = {"key": doc.ents[1]}
    doc.ents[0]._.ent_attr = "ent-0"
    doc.ents[1]._.test_dict = {"key": doc.ents[0]}
    doc.ents[2]._.test_dict = {"key": doc.ents[0]}
    doc.ents[2]._.ent_attr = "ent-2"
    doc.spans["section"] = [doc[:]]
    doc.spans["section"][0]._.ent_attr = "section"
    subdocs = list(eds.split(regex="\n\n")(doc))

    assert len(subdocs) == 2
    assert subdocs[0].text == "This is a test. Another test.\n\n"
    assert subdocs[0]._.global_attr == "global"
    assert subdocs[0].ents[0]._.test_dict["key"] == subdocs[0].ents[1]
    assert subdocs[0].ents[0]._.ent_attr == "ent-0"
    assert subdocs[0].ents[1]._.test_dict["key"] == subdocs[0].ents[0]
    assert subdocs[0].spans["section"][0].text == "This is a test. Another test.\n\n"
    assert subdocs[0].spans["section"][0]._.ent_attr == "section"

    assert subdocs[1].text == "A third test!"
    assert subdocs[1]._.global_attr == "global"
    assert subdocs[1].ents[0]._.test_dict == {}
    assert subdocs[1].ents[0]._.ent_attr == "ent-2"
    assert subdocs[1].spans["section"][0].text == "A third test!"
    assert subdocs[1].spans["section"][0]._.ent_attr == "section"


def test_filter():
    txt = """This is a test. Another test."""
    doc = edsnlp.blank("eds")(txt)
    subdocs = list(
        eds.split(
            regex=r"[.!?]\s+()[A-Z]",
            filter_expr='"Another" in doc.text',
        )(doc)
    )
    assert len(subdocs) == 1
    assert subdocs[0].text == "Another test."


def test_max_length():
    txt = (
        "Le patient mange des pates depuis le début du confinement, "
        "il est donc un peu ballonné, mais pense revenir à un régime plus "
        "équilibré en mangeant des légumes et des fruits."
    )
    doc = edsnlp.blank("eds")(txt)
    texts = [d.text for d in eds.split(max_length=12)(doc)]
    assert texts == [
        "Le patient mange des pates depuis le début du confinement, il ",
        "est donc un peu ballonné, mais pense revenir à un régime ",
        "plus équilibré en mangeant des légumes et des fruits.",
    ]



from typing import Any, Optional

from edsnlp import registry
from edsnlp.metrics import Examples, make_examples, prf


def dependency_parsing_metric(
    examples: Examples,
    filter_expr: Optional[str] = None,
):
    """
    Compute the UAS and LAS scores for dependency parsing.

    Parameters
    ----------
    examples : Examples
        The examples to score, either a tuple of (golds, preds) or a list of
        spacy.training.Example objects
    filter_expr : Optional[str]
        The filter expression to use to filter the documents. The expression
        should be a valid Python test and uses the variable `doc` to refer to
        the reference (gold) document.

    Returns
    -------
    Dict[str, float]
    """
    items = {
        "uas": (set(), set()),
        "las": (set(), set()),
    }
    examples = make_examples(examples)
    if filter_expr is not None:
        filter_fn = eval(f"lambda doc: {filter_expr}")
        examples = [eg for eg in examples if filter_fn(eg.reference)]

    for eg_idx, eg in enumerate(examples):
        for token in eg.reference:
            items["uas"][0].add((eg_idx, token.i, token.head.i))
            items["las"][0].add((eg_idx, token.i, token.head.i, token.dep_))

        for token in eg.predicted:
            items["uas"][1].add((eg_idx, token.i, token.head.i))
            items["las"][1].add((eg_idx, token.i, token.head.i, token.dep_))

    return {name: prf(pred, gold)["f"] for name, (pred, gold) in items.items()}


@registry.metrics.register("eds.dep_parsing")
class DependencyParsingMetric:
    def __init__(self, filter_expr: Optional[str] = None):
        self.filter_expr = filter_expr

    __init__.__doc__ = dependency_parsing_metric.__doc__

    def __call__(self, *examples: Any):
        return dependency_parsing_metric(examples, self.filter_expr)



import datetime
import os
from unittest.mock import Mock, patch

import pytest

try:
    import optuna
except ImportError:
    optuna = None

if optuna is None:
    pytest.skip("optuna not installed", allow_module_level=True)

from confit import Config

from edsnlp.tune import (
    compute_importances,
    compute_n_trials,
    compute_remaining_n_trials_possible,
    compute_time_per_trial,
    is_plotly_install,
    load_config,
    optimize,
    process_results,
    tune,
)


def build_trial(number, value, params, datetime_start, datetime_complete):
    trial = Mock(spec=optuna.trial.FrozenTrial)
    trial.number = number
    trial.value = value
    trial.values = [value]
    trial.params = params
    trial.distributions = {
        "param1": optuna.distributions.FloatDistribution(
            high=0.3,
            log=False,
            low=0.0,
            step=0.05,
        ),
        "param2": optuna.distributions.FloatDistribution(
            high=0.3,
            log=False,
            low=0.0,
            step=0.05,
        ),
    }
    trial.datetime_start = datetime_start
    trial.datetime_complete = datetime_complete
    trial.state = optuna.trial.TrialState.COMPLETE
    trial.system_attrs = {}
    trial.user_attrs = {}
    return trial


@pytest.fixture
def study():
    study = Mock(spec=optuna.study.Study)
    study.study_name = "mock_study"
    study._is_multi_objective.return_value = False

    trials = []
    trial_0 = build_trial(
        number=0,
        value=0.9,
        params={"param1": 0.15, "param2": 0.3},
        datetime_start=datetime.datetime(2025, 1, 1, 12, 0, 0),
        datetime_complete=datetime.datetime(2025, 1, 1, 12, 5, 0),
    )
    trials.append(trial_0)

    trial_1 = build_trial(
        number=1,
        value=0.75,
        params={"param1": 0.05, "param2": 0.2},
        datetime_start=datetime.datetime(2025, 1, 1, 12, 5, 0),
        datetime_complete=datetime.datetime(2025, 1, 1, 12, 10, 0),
    )
    trials.append(trial_1)

    trial_2 = build_trial(
        number=2,
        value=0.99,
        params={"param1": 0.3, "param2": 0.25},
        datetime_start=datetime.datetime(2025, 1, 1, 12, 10, 0),
        datetime_complete=datetime.datetime(2025, 1, 1, 12, 15, 0),
    )
    trials.append(trial_2)

    study.trials = trials
    study.get_trials.return_value = trials
    study.best_trial = trials[2]
    return study


@pytest.mark.parametrize("ema", [True, False])
def test_compute_time_per_trial_with_ema(study, ema):
    result = compute_time_per_trial(study, ema=ema, alpha=0.1)
    assert result == pytest.approx(300.00)


@pytest.mark.parametrize(
    "gpu_hours, time_per_trial, expected_n_trials, raises_exception",
    [
        (1, 120, 30, False),
        (0.5, 3600, None, True),
    ],
)
def test_compute_n_trials(
    gpu_hours, time_per_trial, expected_n_trials, raises_exception
):
    if raises_exception:
        with pytest.raises(ValueError):
            compute_n_trials(gpu_hours, time_per_trial)
    else:
        result = compute_n_trials(gpu_hours, time_per_trial)
        assert result == expected_n_trials


def test_compute_importances(study):
    importance = compute_importances(study)
    assert importance == {"param2": 0.5239814153755754, "param1": 0.4760185846244246}


@pytest.mark.parametrize("viz", [True, False])
@pytest.mark.parametrize(
    "config_path", ["tests/tuning/config.yml", "tests/tuning/config.cfg"]
)
def test_process_results(study, tmpdir, viz, config_path):
    output_dir = tmpdir.mkdir("output")
    config = {
        "train": {
            "param1": None,
        },
        ".lr": {
            "param2": 0.01,
        },
    }
    hyperparameters = {
        "train.param1": {
            "type": "int",
            "alias": "param1",
            "low": 2,
            "high": 8,
            "step": 2,
        },
    }
    best_params, importances = process_results(
        study, output_dir, viz, config, config_path, hyperparameters
    )

    assert isinstance(best_params, dict)
    assert isinstance(importances, dict)

    results_file = os.path.join(output_dir, "results_summary.txt")
    assert os.path.exists(results_file)

    with open(results_file, "r") as f:
        content = f.read()
        assert "Study Summary" in content
        assert "Best trial" in content
        assert "Value" in content
        assert "Params" in content
        assert "Importances" in content

    if config_path.endswith("yml") or config_path.endswith("yaml"):
        config_file = os.path.join(output_dir, "config.yml")
    else:
        config_file = os.path.join(output_dir, "config.cfg")
    assert os.path.exists(config_file), f"Expected file {config_file} not found"

    with open(config_file, "r", encoding="utf-8") as f:
        content = f.read()
    assert (
        "# My usefull comment" in content
    ), f"Expected comment not found in {config_file}"

    if viz:
        optimization_history_file = os.path.join(
            output_dir, "optimization_history.html"
        )
        assert os.path.exists(
            optimization_history_file
        ), f"Expected file {optimization_history_file} not found"

        parallel_coord_file = os.path.join(output_dir, "parallel_coordinate.html")
        assert os.path.exists(
            parallel_coord_file
        ), f"Expected file {parallel_coord_file} not found"

        contour_file = os.path.join(output_dir, "contour.html")
        assert os.path.exists(contour_file), f"Expected file {contour_file} not found"

        edf_file = os.path.join(output_dir, "edf.html")
        assert os.path.exists(edf_file), f"Expected file {edf_file} not found"

        timeline_file = os.path.join(output_dir, "timeline.html")
        assert os.path.exists(timeline_file), f"Expected file {timeline_file} not found"


def test_compute_remaining_n_trials_possible(study):
    gpu_hours = 0.5
    remaining_trials = compute_remaining_n_trials_possible(study, gpu_hours)
    assert remaining_trials == 3


@patch("edsnlp.tune.objective_with_param")
@patch("optuna.study.Study.optimize")
@pytest.mark.parametrize("has_study", [True, False])
def test_optimize(mock_objective_with_param, mock_optimize_study, has_study, study):
    mock_objective_with_param.return_value = 0.9
    metric = ("ner", "micro", "f")
    checkpoint_dir = "./checkpoint"

    if has_study:

        def pass_fn(obj, n_trials, callbacks):
            pass

        study.optimize = pass_fn
        study = optimize(
            "config_path",
            tuned_parameters={},
            n_trials=1,
            metric=metric,
            checkpoint_dir=checkpoint_dir,
            study=study,
        )
        assert isinstance(study, Mock)
        assert len(study.trials) == 3

    else:
        study = optimize(
            "config_path",
            tuned_parameters={},
            n_trials=1,
            metric=metric,
            checkpoint_dir=checkpoint_dir,
            study=None,
        )
        assert isinstance(study, optuna.study.Study)
        assert len(study.trials) == 0


@patch("edsnlp.tune.optimize")
@patch("edsnlp.tune.process_results")
@patch("edsnlp.tune.load_config")
@patch("edsnlp.tune.compute_n_trials")
@patch("edsnlp.tune.update_config")
@pytest.mark.parametrize("n_trials", [10, None])
@pytest.mark.parametrize("two_phase_tuning", [False, True])
def test_tune(
    mock_update_config,
    mock_compute_n_trials,
    mock_load_config,
    mock_process_results,
    mock_optimize,
    study,
    n_trials,
    two_phase_tuning,
):
    mock_load_config.return_value = {"train": {}, "scorer": {}, "val_data": {}}
    mock_update_config.return_value = None, {"train": {}, "scorer": {}, "val_data": {}}
    mock_optimize.return_value = study
    mock_process_results.return_value = ({}, {})
    mock_compute_n_trials.return_value = 10
    config_meta = {"config_path": ["fake_path"]}
    hyperparameters = {
        "param1": {"type": "float", "low": 0.0, "high": 1.0},
        "param2": {"type": "float", "low": 0.0, "high": 1.0},
    }
    output_dir = "output_dir"
    checkpoint_dir = "checkpoint_dir"
    gpu_hours = 0.25
    seed = 42

    tune(
        config_meta=config_meta,
        hyperparameters=hyperparameters,
        output_dir=output_dir,
        checkpoint_dir=checkpoint_dir,
        gpu_hours=gpu_hours,
        n_trials=n_trials,
        two_phase_tuning=two_phase_tuning,
        seed=seed,
    )

    mock_load_config.assert_called_once()

    if two_phase_tuning:
        if n_trials is None:
            assert mock_compute_n_trials.call_count == 2  # 1 at begining + 1 at end
            assert mock_optimize.call_count == 3  # 1 at begining + 2 for tuning
        else:
            mock_compute_n_trials.assert_not_called()
            assert mock_optimize.call_count == 2  # 2 for tuning

        assert mock_process_results.call_count == 2  # one for each phase

    else:
        if n_trials is None:
            assert mock_compute_n_trials.call_count == 2  # 1 at begining + 1 at end
            assert (
                mock_optimize.call_count == 3
            )  # 1 at begining + 1 for tuning + 1 at end
        else:
            mock_compute_n_trials.assert_not_called()
            assert mock_optimize.call_count == 1  # 1 for tuning

        mock_process_results.assert_called_once()


@patch("importlib.util.find_spec")
def test_plotly(mock_importlib_util_find_spec):
    mock_importlib_util_find_spec.return_value = None
    assert not is_plotly_install()


def test_load_config(tmpdir):
    cfg = """\
    "a":
        "aa": 1
    "b": 2
    "c": "test"
    """
    config_dir = tmpdir.mkdir("configs")
    config_path = os.path.join(config_dir, "config.yml")
    Config.from_yaml_str(cfg).to_disk(config_path)
    config = load_config(config_path)
    assert isinstance(config, Config)
    with pytest.raises(FileNotFoundError):
        load_config("wrong_path")
